#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import logging
from typing import Any, Dict, List, Union

import torch
from classy_vision.dataset import ClassyDataset, build_dataset
from classy_vision.generic.distributed_util import (
    all_reduce_mean,
    init_distributed_data_parallel_model,
    is_distributed_training_run,
)
from classy_vision.generic.perf_stats import PerfTimer
from classy_vision.generic.util import (
    copy_model_to_gpu,
    recursive_copy_to_gpu,
    update_classy_state,
)
from classy_vision.losses import ClassyLoss, build_loss
from classy_vision.meters import build_meters
from classy_vision.models import ClassyModel, build_model
from classy_vision.optim import ClassyOptimizer, build_optimizer

from . import register_task
from .classy_task import ClassyTask


@register_task("classification_task")
class ClassificationTask(ClassyTask):
    """Basic classification training task.

    This task encapsultates all of the components and steps needed to
    train a classifier using a :obj:`ClassyTrainer`.

    Assumes a train / test phase per each epoch and that the datasets
    have the same API as the map-style Dataset class in
    torch.utils.data.dataset (in particular, this task makes use of
    the len).  If you are using an IterableDataset then a custom task
    may be appropriate.

    Attributes:
        loss: Loss (see :obj:`ClassyLoss`) function used for computing the loss in
            each forward pass
        datasets: Mapping from a phase_type in ["train", "test']
            to dataset used for training (or testing)
        meters: List of meters (see :obj:`ClassyMeter`) to calculate during training
        num_epochs: Number of epochs (passes over dataset) to train
        test_only: Used to only run the test phase
        base_model: Model to be trained, unwrapped in DDP or DP wrappers
        optimizer: Optimizer used in train step
        checkpoint: Serializable dict which represents state in training
        phases: List of phase specific information, e.g. if phase is
            train / test.
        hooks: List of hooks to apply during training
        train: Phase type, if true it means we are training,
            false means testing
        distributed_model: Base model, but wrapped in DDP (DistributedDataParallel)
        phase_idx: Current phase id, first phase is 0, if task has not started
            training then returns -1
        train_phase_idx: Only counts train phases
        num_updates: Number of total parameter updates applied to model
            by the optimizer
        data_iterator: Iterator which can be used to obtain batches
        num_samples_this_phase: Number of samples ran this phase
        losses: Loss curve

    """

    def __init__(self):
        """Constructs a ClassificationTask
        """
        super().__init__()

        self.loss = None
        self.datasets = {}
        self.meters = []
        self.num_epochs = 1
        self.test_only = False
        self.base_model = None
        self.optimizer = None
        self.checkpoint = None
        self.phases = []
        self.hooks = []
        self.train = True
        self.distributed_model = None
        self.phase_idx = -1
        self.train_phase_idx = -1
        self.num_updates = 0
        self.data_iterator = None
        self.num_samples_this_phase = 0
        self.losses = []

    def set_checkpoint(self, checkpoint):
        """Sets checkpoint on task.

        Args:
            checkpoint: A serializable dict representing current task state
        """
        assert (
            checkpoint is None or "classy_state_dict" in checkpoint
        ), "Checkpoint does not contain classy_state_dict"
        self.checkpoint = checkpoint

    def set_num_epochs(self, num_epochs: Union[int, float]):
        """Set number of epochs to be run.

        Args:
           num_epochs: Number of epochs to run task
        """
        self.num_epochs = num_epochs
        return self

    def set_dataset(self, dataset: ClassyDataset, phase_type: str):
        """Set dataset for phase type on task

        Args:
            dataset: ClassyDataset for returning samples.
            phase_type: str must be one of "train" or "test"
        """
        assert phase_type in [
            "train",
            "test",
        ], "phase_type must be in ['train', 'test']"
        self.datasets[phase_type] = dataset
        return self

    def set_optimizer(self, optimizer: ClassyOptimizer):
        """Set optimizer for task

        Args:
            optimizer: optimizer for task
        """
        self.optimizer = optimizer
        return self

    def set_loss(self, loss: ClassyLoss):
        """Set loss function for task

        Args:
            loss: loss for task
        """
        self.loss = loss
        return self

    def set_meters(self, meters: List["ClassyMeter"]):
        """Set meters for task

        Args:
            meters: list of meters to compute during training
        """
        self.meters = meters
        return self

    def set_hooks(self, hooks: List["ClassyHook"]):
        """Set hooks for task

        Args:
            hooks: List of hooks to apply during training
        """
        from classy_vision.hooks import ClassyHook

        assert isinstance(hooks, list)
        assert all(isinstance(hook, ClassyHook) for hook in hooks)
        assert len({hook.name() for hook in hooks}) == len(
            hooks
        ), "Cannot have repeated hooks of the same class"

        self.hooks = hooks
        return self

    def set_model(self, model: ClassyModel):
        """Set model for task

        Args:
            model: Model to be trained
        """
        self.base_model = model
        return self

    def set_test_only(self, test_only: bool):
        """Set test only flag

        Args:
            test_only: If true, only test phases will be run
        """
        self.test_only = test_only
        return self

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassificationTask":
        """Instantiates a ClassificationTask from a configuration.

        Args:
            config: A configuration for a ClassificationTask.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A ClassificationTask instance.
        """
        optimizer_config = config["optimizer"]
        optimizer_config["num_epochs"] = config["num_epochs"]

        datasets = {}
        phase_types = ["train", "test"]
        for phase_type in phase_types:
            datasets[phase_type] = build_dataset(config["dataset"][phase_type])
        loss = build_loss(config["loss"])
        test_only = config.get("test_only", False)
        meters = build_meters(config.get("meters", {}))
        model = build_model(config["model"])
        # put model in eval mode in case any hooks modify model states, it'll
        # be reset to train mode before training
        model.eval()
        optimizer = build_optimizer(optimizer_config)

        task = (
            cls()
            .set_num_epochs(config["num_epochs"])
            .set_loss(loss)
            .set_test_only(test_only)
            .set_model(model)
            .set_optimizer(optimizer)
            .set_meters(meters)
        )
        for phase_type in phase_types:
            task.set_dataset(datasets[phase_type], phase_type)

        return task

    @property
    def num_batches_per_phase(self):
        """Returns number of batches in current phase iterator
        """
        return len(self.data_iterator)

    @property
    def model(self):
        """Returns model used in training (can be wrapped with DDP)
        """
        return (
            self.distributed_model if is_distributed_training_run() else self.base_model
        )

    @property
    def phase_type(self):
        """Returns current phase type. String with value "train" or "test"
        """
        return "train" if self.train else "test"

    @property
    def eval_phase_idx(self):
        """Returns current evaluation phase
        """
        return self.phase_idx - self.train_phase_idx - 1

    def get_data_iterator(self):
        """Returns data iterator for current phase
        """
        return self.data_iterator

    def get_total_training_phases(self):
        """
        Returns the total number of "train" phases in the task
        """
        num_training_phases = 0
        for phase in self.phases:
            if phase["train"] is True:
                num_training_phases += 1
        return num_training_phases

    def _build_phases(self):
        """Returns list of phases from config.

        These phases will look like:
        {
          train: is this a train or test phase?
          optimizer: optimizer settings
        }

        If this is a test only run, then only test phases will be
        generated, if this is a training run, then x phases = x train
        phases + x test phases, interleaved.
        """
        if not self.test_only:
            phases = [{"train": True} for _ in range(self.num_epochs)]

            final_phases = []
            for phase in phases:
                final_phases.append(phase)
                final_phases.append({"train": False})
            return final_phases

        return [{"train": False} for _ in range(self.num_epochs)]

    def build_dataloader(
        self,
        phase_type,
        num_workers,
        pin_memory,
        multiprocessing_context=None,
        **kwargs,
    ):
        """Buildss a dataloader iterable for a particular phase type.

        Args:
            phase_type: "train" or "test" iterable
            num_workers: Number of dataloading processes. If 0,
                dataloading is done on main process. See PyTorch dataloader
                documentation for more details on num_workers and the usage
                of python multiprocessing in dataloaders
            pin_memory: if true pin memory on GPU. See PyTorch dataloader
                documentation for details on pin_memory.
            multiprocessing_context: Determines how processes are spawned.
                Value must be one of None, "spawn", "fork", "forkserver".
                If None, then context is inherited from parent process

        Returns:
            Returns a iterable over the dataset
        """
        return self.datasets[phase_type].iterator(
            num_workers=num_workers,
            pin_memory=pin_memory,
            multiprocessing_context=multiprocessing_context,
            **kwargs,
        )

    def build_dataloaders(
        self, num_workers, pin_memory, multiprocessing_context=None, **kwargs
    ):
        """Build a dataloader for each phase type

        Args:
            num_workers: Number of dataloading processes. If 0,
                dataloading is done on main process. See PyTorch dataloader
                documentation for more details on num_workers and the usage
                of python multiprocessing in dataloaders
            pin_memory: if true pin memory on GPU. See PyTorch dataloader
                documentation for details on pin_memory.
            multiprocessing_context: Determines how processes are spawned.
                Value must be one of None, "spawn", "fork", "forkserver".
                If None, then context is inherited from parent process

        Returns:
            Returns an iterable over the dataset associated with each phase_type
        """
        return {
            phase_type: self.build_dataloader(
                phase_type,
                num_workers=num_workers,
                pin_memory=pin_memory,
                multiprocessing_context=multiprocessing_context,
                **kwargs,
            )
            for phase_type in self.datasets.keys()
        }

    def prepare(
        self,
        num_dataloader_workers=0,
        pin_memory=False,
        use_gpu=False,
        dataloader_mp_context=None,
    ):
        """Prepares task for training, populates all derived attributes

        Args:
            num_dataloader_workers: Number of dataloading processes. If 0,
                dataloading is done on main process
            pin_memory: if true pin memory on GPU
            use_gpu: if true, load model, optimizer, loss, etc on GPU
            dataloader_mp_context: Determines how processes are spawned.
                Value must be one of None, "spawn", "fork", "forkserver".
                If None, then context is inherited from parent process
        """
        self.phases = self._build_phases()
        self.dataloaders = self.build_dataloaders(
            num_workers=num_dataloader_workers,
            pin_memory=pin_memory,
            multiprocessing_context=dataloader_mp_context,
        )

        # move the model and loss to the right device
        if use_gpu:
            self.loss.cuda()
            self.base_model = copy_model_to_gpu(self.base_model)
        else:
            self.loss.cpu()
            self.base_model.cpu()

        # initialize the pytorch optimizer now since the model has been moved to
        # the appropriate device
        self.optimizer.init_pytorch_optimizer(self.base_model)

        classy_state_dict = (
            None
            if self.checkpoint is None
            else self.checkpoint.get("classy_state_dict")
        )

        if classy_state_dict is not None:
            state_load_success = update_classy_state(self, classy_state_dict)
            assert (
                state_load_success
            ), "Update classy state from checkpoint was unsuccessful."

    def init_distributed_data_parallel_model(self):
        """Sets up distributed dataparallel and wraps model in DDP
        """
        assert (
            self.distributed_model is None
        ), "init_ddp_non_elastic must only be called once"

        self.distributed_model = init_distributed_data_parallel_model(self.base_model)

    @property
    def where(self):
        """Returns the proportion of training that has completed.

        Returned value is a float in the range [0, 1)
        """
        current_step = self.num_updates / self.get_global_batchsize()
        num_steps = self.get_total_training_phases() * self.num_batches_per_phase
        where = current_step / num_steps

        assert where >= 0 and where < 1, f"Invalid where: {where}"

        return where

    def get_classy_state(self, deep_copy: bool = False):
        """Returns serialiable state of task

        Args:
            deep_copy: If true, does a deep copy of state before returning.
        """
        classy_state_dict = {
            "train": self.train,
            "base_model": self.base_model.get_classy_state(),
            "meters": [meter.get_classy_state() for meter in self.meters],
            "optimizer": self.optimizer.get_classy_state(),
            "phase_idx": self.phase_idx,
            "train_phase_idx": self.train_phase_idx,
            "num_updates": self.num_updates,
            "num_samples_this_phase": self.num_samples_this_phase,
            "losses": self.losses,
            "hooks": {hook.name(): hook.get_classy_state() for hook in self.hooks},
        }
        if deep_copy:
            classy_state_dict = copy.deepcopy(classy_state_dict)
        return classy_state_dict

    def set_classy_state(self, state):
        """Set task state

        Args:
            state: Dict containing state of a task
        """
        self.train = state["train"]
        self.base_model.set_classy_state(state["base_model"])
        for meter, meter_state in zip(self.meters, state["meters"]):
            meter.set_classy_state(meter_state)
        self.optimizer.set_classy_state(state["optimizer"])
        self.phase_idx = state["phase_idx"]
        self.train_phase_idx = state["train_phase_idx"]
        self.num_updates = state["num_updates"]
        self.num_samples_this_phase = state["num_samples_this_phase"]
        self.losses = state["losses"]
        for hook in self.hooks:
            # we still want to be able to run when new hooks are added or old
            # hooks are removed
            if hook.name() in state["hooks"]:
                hook.set_classy_state(state["hooks"][hook.name()])
            else:
                logging.warn(f"No state found for hook: {hook.name()}")
        # TODO (mannatsingh): Figure out how to set the state of the dataloaders
        # Re-build dataloader & re-create iterator.
        self._recreate_data_loader_from_dataset()
        self._reshuffle_data()
        self.create_data_iterator()
        # Set up pytorch module in train vs eval mode, update optimizer.
        self._set_model_train_mode()

    def train_step(self, use_gpu, local_variables=None):
        """Train step to be executed in train loop

        Args:
            use_gpu: if true, execute training on GPU
            local_variables: Dict containing intermediate values
                in train_step for access by hooks
        """
        from classy_vision.hooks import ClassyHookFunctions

        if local_variables is None:
            local_variables = {}

        # We'll time train_step and some of its sections, and accumulate values
        # into perf_stats if it were defined in local_variables:
        perf_stats = local_variables.get("perf_stats", None)
        timer_train_step = PerfTimer("train_step_total", perf_stats)
        timer_train_step.start()

        # Process next sample
        with PerfTimer("read_sample", perf_stats):
            sample = next(self.get_data_iterator())
            local_variables["sample"] = sample

            assert (
                isinstance(local_variables["sample"], dict)
                and "input" in local_variables["sample"]
                and "target" in local_variables["sample"]
            ), (
                f"Returned sample [{sample}] is not a map with 'input' and"
                + "'target' keys"
            )

        self.run_hooks(local_variables, ClassyHookFunctions.on_sample.name)

        # Copy sample to GPU
        local_variables["target"] = local_variables["sample"]["target"]
        if use_gpu:
            for key, value in local_variables["sample"].items():
                local_variables["sample"][key] = recursive_copy_to_gpu(
                    value, non_blocking=True
                )

        # Only need gradients during training
        context = torch.enable_grad() if self.train else torch.no_grad()
        with context:
            # Forward pass
            with PerfTimer("forward", perf_stats):
                local_variables["output"] = self.model(
                    local_variables["sample"]["input"]
                )

            self.run_hooks(local_variables, ClassyHookFunctions.on_forward.name)

            model_output = local_variables["output"]
            target = local_variables["sample"]["target"]
            local_variables["local_loss"] = self.loss(model_output, target)

            # NOTE: This performs an all_reduce_mean() on the losses across the
            # replicas.  The reduce should ideally be weighted by the length of
            # the targets on each replica. This will only be an issue when
            # there are dummy samples present (once an epoch) and will only
            # impact the loss reporting (slightly).
            with PerfTimer("loss_allreduce", perf_stats):
                local_variables["loss"] = local_variables["local_loss"].detach().clone()
                local_variables["loss"] = all_reduce_mean(local_variables["loss"])

            self.losses.append(
                local_variables["loss"].data.cpu().item()
                * local_variables["target"].size(0)
            )

            model_output_cpu = model_output.cpu() if use_gpu else model_output

            # Update meters
            with PerfTimer("meters_update", perf_stats):
                for meter in self.meters:
                    meter.update(
                        model_output_cpu, target.detach().cpu(), is_train=self.train
                    )
            # After both loss and meters are updated, we run hooks. Among hooks,
            # `LossLrMeterLoggingHook` will log both loss and meter status
            self.run_hooks(local_variables, ClassyHookFunctions.on_loss_and_meter.name)

        num_samples_in_step = self.get_global_batchsize()
        self.num_samples_this_phase += num_samples_in_step

        # For training phases, run backwards pass / update optimizer
        if self.train:
            with PerfTimer("backward", perf_stats):
                self.optimizer.backward(local_variables["local_loss"])

            self.run_hooks(local_variables, ClassyHookFunctions.on_backward.name)

            self.optimizer.update_schedule_on_step(self.where)
            with PerfTimer("optimizer_step", perf_stats):
                self.optimizer.step()

            self.run_hooks(local_variables, ClassyHookFunctions.on_update.name)

            self.num_updates += num_samples_in_step

        timer_train_step.stop()
        timer_train_step.record()

    def advance_phase(self):
        """Performs bookkeeping / task updates between phases

        Increments phase idx, resets meters, resets loss history,
        resets counters, shuffles dataset, rebuilds iterators, and
        sets the train / test state for phase.
        """
        logging.info("Advancing phase")
        # Reset meters for next phase / epoch
        for meter in self.meters:
            meter.reset()

        # Reset loss history for next epoch
        self.losses = []

        # Setup new phase
        self.num_samples_this_phase = 0
        self.phase_idx += 1
        phase = self.phases[self.phase_idx]
        self.train = True if phase["train"] else False
        if self.train:
            self.train_phase_idx += 1

        # Re-build dataloader & re-create iterator anytime membership changes.
        self._recreate_data_loader_from_dataset()
        self._reshuffle_data()
        self.create_data_iterator()
        # Set up pytorch module in train vs eval mode, update optimizer.
        self._set_model_train_mode()

    def done_training(self):
        """Stop condition for training
        """
        return self.phase_idx + 1 >= len(self.phases)

    def _recreate_data_loader_from_dataset(self, phase_type=None):
        """
        This utility is invoked to re-create the data loader object
        for the current phase of execution, using the existing dataset.
        This is sufficient when advancing phases.
        """
        if phase_type is None:
            phase_type = self.phase_type

        logging.info("Recreating data loader for new phase")
        num_workers = 0
        if hasattr(self.dataloaders[phase_type], "num_workers"):
            num_workers = self.dataloaders[phase_type].num_workers
        pin_memory = False
        if hasattr(self.dataloaders[phase_type], "pin_memory"):
            pin_memory = self.dataloaders[phase_type].pin_memory
        multiprocessing_context = None
        if hasattr(self.dataloaders[phase_type], "multiprocessing_context"):
            multiprocessing_context = self.dataloaders[
                phase_type
            ].multiprocessing_context
        if phase_type == "test":
            current_phase_id = 0
        else:
            current_phase_id = max(self.train_phase_idx, 0)

        self.dataloaders[phase_type] = self.build_dataloader(
            phase_type=phase_type,
            num_workers=num_workers,
            pin_memory=pin_memory,
            multiprocessing_context=multiprocessing_context,
            current_phase_id=current_phase_id,
        )

    def _reshuffle_data(self):
        """Shuffles the dataset if needed.
        """
        if hasattr(self.dataloaders[self.phase_type].dataset, "do_shuffle"):
            self.dataloaders[self.phase_type].dataset.do_shuffle(
                epoch_num=self.phase_idx
            )
            logging.info("Data shuffled.")

    def create_data_iterator(self):
        """Creates data iterator for phase.
        """
        # Delete iterator explicitly so that all dataloader processes
        # are cleaned up.
        del self.data_iterator
        self.data_iterator = iter(self.dataloaders[self.phase_type])

    def _set_model_train_mode(self):
        """Set train mode for model
        """
        phase = self.phases[self.phase_idx]
        self.base_model.train(phase["train"])

        if self.train and self.train_phase_idx >= 0:
            self.optimizer.update_schedule_on_epoch(self.where)

    # TODO: Functions below should be better abstracted into the dataloader
    # abstraction
    def get_batchsize_per_replica(self):
        """Return local replica's batchsize for dataset (e.g. batchsize per GPU)
        """
        # TODO(T47573564) - cleaner abstraction
        return self.dataloaders[self.phase_type].dataset.get_batchsize_per_replica()

    def get_global_batchsize(self):
        """Return global batchsize across all trainers
        """
        return self.dataloaders[self.phase_type].dataset.get_global_batchsize()

    def get_total_samples_trained_this_phase(self):
        """Returns the total number of samples processed in current phase
        """
        # TODO(T47573564) - cleaner abstraction
        # TODO(T47387605) - instead of get_world_size, we need the max world
        # size for elasticity to match parity with Uru and other systems,
        # although DPP will solve this by dynamically re-sharding.
        return self.num_samples_this_phase



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict

from classy_vision import tasks
from classy_vision.generic.profiler import profile, summarize_profiler_info
from classy_vision.hooks.classy_hook import ClassyHook


class ProfilerHook(ClassyHook):
    """
    Hook to profile a model and to show model runtime information, such as
        the time breakdown in milliseconds of forward/backward pass.
    """

    on_rendezvous = ClassyHook._noop
    on_phase_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_update = ClassyHook._noop
    on_phase_end = ClassyHook._noop
    on_end = ClassyHook._noop

    def on_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Profile the forward pass."""
        logging.info("Profiling forward pass...")
        batchsize_per_replica = getattr(
            task.dataloaders[task.phase_type].dataset, "batchsize_per_replica", 1
        )
        input_shape = task.base_model.input_shape
        p = profile(
            task.model,
            batchsize_per_replica=batchsize_per_replica,
            input_shape=input_shape,
            input_key=getattr(task.base_model, "input_key", None),
        )
        logging.info(summarize_profiler_info(p))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import unittest.mock as mock
from itertools import product
from test.generic.config_utils import get_test_mlp_task_config, get_test_task_config

from classy_vision.hooks import LossLrMeterLoggingHook
from classy_vision.optim.param_scheduler import UpdateInterval
from classy_vision.tasks import ClassyTask, build_task
from classy_vision.trainer import LocalTrainer


class TestLossLrMeterLoggingHook(unittest.TestCase):
    @mock.patch("classy_vision.hooks.loss_lr_meter_logging_hook.get_rank")
    def test_logging(self, mock_get_rank: mock.MagicMock) -> None:
        """
        Test that the logging happens as expected and the loss and lr values are
        correct.
        """
        rank = 5
        mock_get_rank.return_value = rank

        # set up the task and state
        config = get_test_task_config()
        config["dataset"]["train"]["batchsize_per_replica"] = 2
        config["dataset"]["test"]["batchsize_per_replica"] = 5
        task = build_task(config)
        task.prepare()

        losses = [1.2, 2.3, 3.4, 4.5]

        local_variables = {}
        task.phase_idx = 0

        loss_vals = {"train": 1.425, "test": 0.57}

        for log_freq, phase_type in product([5, None], loss_vals):
            task.train = phase_type == "train"

            # create a loss lr meter hook
            loss_lr_meter_hook = LossLrMeterLoggingHook(log_freq=log_freq)

            # check that _log_loss_meters() is called after on_loss_and_meter() every
            # log_freq batches and after on_phase_end()
            # and _log_lr() is called after on_update() every log_freq batches
            # and after on_phase_end()
            with mock.patch.object(loss_lr_meter_hook, "_log_loss_meters") as mock_fn:
                with mock.patch.object(loss_lr_meter_hook, "_log_lr") as mock_lr_fn:
                    num_batches = 20

                    for i in range(num_batches):
                        task.losses = list(range(i))
                        loss_lr_meter_hook.on_loss_and_meter(task, local_variables)
                        loss_lr_meter_hook.on_update(task, local_variables)
                        if log_freq is not None and i and i % log_freq == 0:
                            mock_fn.assert_called_with(task, local_variables)
                            mock_fn.reset_mock()
                            mock_lr_fn.assert_called_with(task, local_variables)
                            mock_lr_fn.reset_mock()
                            continue
                        mock_fn.assert_not_called()
                        mock_lr_fn.assert_not_called()

                    loss_lr_meter_hook.on_phase_end(task, local_variables)
                    mock_fn.assert_called_with(task, local_variables)
                    if task.train:
                        mock_lr_fn.assert_called_with(task, local_variables)

            # test _log_loss_lr_meters()
            task.losses = losses

            with self.assertLogs():
                loss_lr_meter_hook._log_loss_meters(task, local_variables)
                loss_lr_meter_hook._log_lr(task, local_variables)

            task.phase_idx += 1

    def test_logged_lr(self):
        # Mock LR scheduler
        def scheduler_mock(where):
            return where

        mock_lr_scheduler = mock.Mock(side_effect=scheduler_mock)
        mock_lr_scheduler.update_interval = UpdateInterval.STEP
        config = get_test_mlp_task_config()
        config["num_epochs"] = 3
        config["dataset"]["train"]["batchsize_per_replica"] = 5
        config["dataset"]["test"]["batchsize_per_replica"] = 5
        task = build_task(config)
        task.optimizer.lr_scheduler = mock_lr_scheduler
        trainer = LocalTrainer()

        # 2 LR updates per epoch
        # At end of each epoch for train, LR is logged an additional time
        lr_order = [0.0, 1 / 6, 1 / 6, 2 / 6, 3 / 6, 3 / 6, 4 / 6, 5 / 6, 5 / 6]
        lr_list = []

        def mock_log_lr(task: ClassyTask, local_variables) -> None:
            lr_list.append(task.optimizer.lr)

        with mock.patch.object(
            LossLrMeterLoggingHook, "_log_lr", side_effect=mock_log_lr
        ):
            hook = LossLrMeterLoggingHook(1)
            task.set_hooks([hook])
            trainer.train(task)
            self.assertEqual(lr_list, lr_order)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import json
import os
import subprocess
import sys
import tempfile
import unittest
from pathlib import Path
from test.generic.config_utils import get_test_mlp_task_config

import torch


class TestDistributedTrainer(unittest.TestCase):
    def setUp(self):
        config = get_test_mlp_task_config()
        invalid_config = copy.deepcopy(config)
        invalid_config["name"] = "invalid_task"
        self.config_files = {}
        for config_key, config in [
            ("config", config),
            ("invalid_config", invalid_config),
        ]:
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
                json.dump(config, f)
                f.flush()
                self.config_files[config_key] = f.name
        self.path = Path(__file__).parent.absolute()

    def tearDown(self):
        for config_file in self.config_files.values():
            os.unlink(config_file)

    def test_training(self):
        """Checks we can train a small MLP model."""

        num_processes = 2
        device = "gpu" if torch.cuda.is_available() else "cpu"

        for config_key, expected_success in [
            ("invalid_config", False),
            ("config", True),
        ]:
            cmd = f"""{sys.executable} -m torch.distributed.launch \
            --nnodes=1 \
            --nproc_per_node={num_processes} \
            --master_addr=127.0.0.1 \
            --master_port=29500 \
            --use_env \
            {self.path}/../classy_train.py \
            --device={device} \
            --config={self.config_files[config_key]} \
            --num_workers=4 \
            --log_freq=100
            """
            result = subprocess.run(cmd, shell=True)
            success = result.returncode == 0
            self.assertEqual(success, expected_success)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import tempfile
from shutil import copy2, move
from typing import Any, Collection, Dict, Optional

from classy_vision import tasks
from classy_vision.generic.distributed_util import is_master
from classy_vision.generic.util import get_checkpoint_dict, save_checkpoint
from classy_vision.hooks.classy_hook import ClassyHook


class CheckpointHook(ClassyHook):
    """
    Hook to checkpoint a model's task.

    Saves the checkpoints in checkpoint_folder.
    """

    on_rendezvous = ClassyHook._noop
    on_phase_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_update = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(
        self,
        checkpoint_folder: str,
        input_args: Any,
        phase_types: Optional[Collection[str]] = None,
        checkpoint_period: int = 1,
    ) -> None:
        """The constructor method of CheckpointHook.

        Args:
            checkpoint_folder - Folder to store checkpoints in
            input_args - Any arguments to save about the runtime setup. For example,
                it is useful to store the config that was used to instantiate the model.
            phase_types - If phase_types is specified, only checkpoint on those phase
                types. Each item in phase_types must be either "train" or "test".
            checkpoint_period - Checkpoint at the end of every x phases (default 1)

        """
        super().__init__()
        self.checkpoint_folder: str = checkpoint_folder
        self.input_args: Any = input_args
        if phase_types is None:
            phase_types = ["train", "test"]
        assert len(phase_types) > 0 and all(
            phase_type in ["train", "test"] for phase_type in phase_types
        ), "phase_types should contain one or more of ['train', 'test']"
        assert (
            isinstance(checkpoint_period, int) and checkpoint_period > 0
        ), "checkpoint period must be positive"

        self.phase_types: Collection[str] = phase_types
        self.checkpoint_period: int = checkpoint_period
        self.phase_counter: int = 0

    def _save_checkpoint(self, task, filename):
        if getattr(task, "test_only", False):
            return
        assert os.path.exists(
            self.checkpoint_folder
        ), "Checkpoint folder '{}' deleted unexpectedly".format(self.checkpoint_folder)

        # save checkpoint:
        logging.info("Saving checkpoint to '{}'...".format(self.checkpoint_folder))
        checkpoint_file = save_checkpoint(
            self.checkpoint_folder, get_checkpoint_dict(task, self.input_args)
        )

        # make copy of checkpoint that won't be overwritten:
        if checkpoint_file:
            tmp_dir = tempfile.mkdtemp()
            tmp_file = os.path.join(tmp_dir, filename)
            copy2(checkpoint_file, tmp_file)
            move(tmp_file, os.path.join(self.checkpoint_folder, filename))

    def on_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        if getattr(task, "test_only", False):
            return
        if not os.path.exists(self.checkpoint_folder):
            err_msg = "Checkpoint folder '{}' does not exist.".format(
                self.checkpoint_folder
            )
            raise FileNotFoundError(err_msg)

    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Checkpoint the task every checkpoint_period phases.

        We do not necessarily checkpoint the task at the end of every phase.
        """
        if not is_master() or task.phase_type not in self.phase_types:
            return

        self.phase_counter += 1
        if self.phase_counter % self.checkpoint_period != 0:
            return

        checkpoint_name = "model_phase-{phase}_end.torch".format(phase=task.phase_idx)
        self._save_checkpoint(task, checkpoint_name)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.step_with_fixed_gamma_scheduler import (
    StepWithFixedGammaParamScheduler,
)


class TestStepWithFixedGammaScheduler(unittest.TestCase):
    _num_epochs = 12

    def _get_valid_config(self):
        return {
            "name": "step_with_fixed_gamma",
            "base_lr": 1,
            "gamma": 0.1,
            "num_decays": 3,
            "num_epochs": self._num_epochs,
        }

    def test_invalid_config(self):
        config = self._get_valid_config()

        # Invalid num epochs
        bad_config = copy.deepcopy(config)
        bad_config["num_epochs"] = -1
        with self.assertRaises(AssertionError):
            StepWithFixedGammaParamScheduler.from_config(bad_config)

        # Invalid num_decays
        bad_config["num_decays"] = 0
        with self.assertRaises(AssertionError):
            StepWithFixedGammaParamScheduler.from_config(bad_config)

        # Invalid base_lr
        bad_config = copy.deepcopy(config)
        bad_config["base_lr"] = -0.01
        with self.assertRaises(AssertionError):
            StepWithFixedGammaParamScheduler.from_config(bad_config)

        # Invalid gamma
        bad_config = copy.deepcopy(config)
        bad_config["gamma"] = [2]
        with self.assertRaises(AssertionError):
            StepWithFixedGammaParamScheduler.from_config(bad_config)

    def test_scheduler(self):
        config = self._get_valid_config()

        scheduler = StepWithFixedGammaParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            1,
            1,
            1,
            0.1,
            0.1,
            0.1,
            0.01,
            0.01,
            0.01,
            0.001,
            0.001,
            0.001,
        ]

        for param, expected_param in zip(schedule, expected_schedule):
            self.assertAlmostEqual(param, expected_param)

    def test_build_step_with_fixed_gamma_scheduler(self):
        config = self._get_valid_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, StepWithFixedGammaParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules


FILE_ROOT = Path(__file__).parent

# Automatically import any Python files in the models/ directory
import_all_modules(FILE_ROOT, "models")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import shutil
import tempfile
import unittest
from test.generic.config_utils import get_test_task_config

from classy_vision.generic.util import load_checkpoint
from classy_vision.hooks import CheckpointHook
from classy_vision.tasks import build_task


class TestCheckpointHook(unittest.TestCase):
    def setUp(self) -> None:
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.base_dir)

    def test_state_checkpointing(self) -> None:
        """
        Test that the state gets checkpointed without any errors, but only on the
        right phase_type and only if the checkpoint directory exists.
        """
        config = get_test_task_config()
        task = build_task(config)
        task.prepare()

        local_variables = {}
        checkpoint_folder = self.base_dir + "/checkpoint_end_test/"
        device = "cpu"
        input_args = {"foo": "bar"}

        # create a checkpoint hook
        checkpoint_hook = CheckpointHook(
            checkpoint_folder, input_args, phase_types=["train"]
        )

        # checkpoint directory doesn't exist
        # call the on start function
        with self.assertRaises(FileNotFoundError):
            checkpoint_hook.on_start(task, local_variables)
        # call the on end phase function
        with self.assertRaises(AssertionError):
            checkpoint_hook.on_phase_end(task, local_variables)
        # try loading a non-existent checkpoint
        checkpoint = load_checkpoint(checkpoint_folder, device)
        self.assertIsNone(checkpoint)

        # create checkpoint dir, verify on_start hook runs
        os.mkdir(checkpoint_folder)
        checkpoint_hook.on_start(task, local_variables)

        # Phase_type is test, expect no checkpoint
        task.train = False
        # call the on end phase function
        checkpoint_hook.on_phase_end(task, local_variables)
        checkpoint = load_checkpoint(checkpoint_folder, device)
        self.assertIsNone(checkpoint)

        task.train = True
        # call the on end phase function
        checkpoint_hook.on_phase_end(task, local_variables)
        # model should be checkpointed. load and compare
        checkpoint = load_checkpoint(checkpoint_folder, device)
        self.assertIsNotNone(checkpoint)
        for key in ["input_args", "classy_state_dict"]:
            self.assertIn(key, checkpoint)
        # not testing for equality of classy_state_dict, that is tested in
        # a separate test
        self.assertDictEqual(checkpoint["input_args"], input_args)

    def test_checkpoint_period(self) -> None:
        """
        Test that the checkpoint_period works as expected.
        """
        config = get_test_task_config()
        task = build_task(config)
        task.prepare()

        local_variables = {}
        checkpoint_folder = self.base_dir + "/checkpoint_end_test/"
        device = "cpu"
        checkpoint_period = 10

        for phase_types in [["train"], ["train", "test"]]:
            # create a checkpoint hook
            checkpoint_hook = CheckpointHook(
                checkpoint_folder,
                {},
                phase_types=phase_types,
                checkpoint_period=checkpoint_period,
            )

            # create checkpoint dir
            os.mkdir(checkpoint_folder)

            # call the on start function
            checkpoint_hook.on_start(task, local_variables)

            # shouldn't create any checkpoints until there are checkpoint_period
            # phases which are in phase_types
            count = 0
            valid_phase_count = 0
            while valid_phase_count < checkpoint_period - 1:
                task.train = count % 2 == 0
                # call the on end phase function
                checkpoint_hook.on_phase_end(task, local_variables)
                checkpoint = load_checkpoint(checkpoint_folder, device)
                self.assertIsNone(checkpoint)
                valid_phase_count += 1 if task.phase_type in phase_types else 0
                count += 1

            # create a phase which is in phase_types
            task.train = True
            # call the on end phase function
            checkpoint_hook.on_phase_end(task, local_variables)
            # model should be checkpointed. load and compare
            checkpoint = load_checkpoint(checkpoint_folder, device)
            self.assertIsNotNone(checkpoint)
            # delete the checkpoint dir
            shutil.rmtree(checkpoint_folder)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from classy_vision.generic.opts import check_generic_args, parse_train_arguments
from classy_vision.generic.util import load_json


try:
    import hydra

    hydra_available = True
except ImportError:
    hydra_available = False

args = None
config = None


if hydra_available:

    @hydra.main(config_path="../hydra/args.yaml")
    def _parse_hydra_args(cfg):
        # This need to be a separate function which sets globals because hydra doesn't
        # support returning from its main function
        global args, config
        args = cfg
        check_generic_args(args)
        config = args.config.to_container()


def parse_args():
    """Parse arguments.

    Parses the args from argparse. If hydra is installed, uses hydra based args
    (experimental).
    """
    if hydra_available:
        global args, config
        _parse_hydra_args()
        return args, config
    else:
        args = parse_train_arguments()
        config = load_json(args.config_file)
        return args, config



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from test.generic.config_utils import get_test_mlp_task_config

from classy_vision.dataset import build_dataset
from classy_vision.hooks import LossLrMeterLoggingHook
from classy_vision.losses import build_loss
from classy_vision.meters import AccuracyMeter
from classy_vision.models import build_model
from classy_vision.optim import build_optimizer
from classy_vision.tasks import ClassificationTask
from classy_vision.trainer import LocalTrainer


class TestLocalTrainer(unittest.TestCase):
    def test_training(self):
        """Checks we can train a small MLP model."""
        config = get_test_mlp_task_config()
        task = (
            ClassificationTask()
            .set_num_epochs(10)
            .set_loss(build_loss(config["loss"]))
            .set_model(build_model(config["model"]))
            .set_optimizer(build_optimizer(config["optimizer"]))
            .set_meters([AccuracyMeter(topk=[1])])
            .set_hooks([LossLrMeterLoggingHook()])
        )
        for split in ["train", "test"]:
            dataset = build_dataset(config["dataset"][split])
            task.set_dataset(dataset, split)

        self.assertTrue(task is not None)

        trainer = LocalTrainer()
        trainer.train(task)
        accuracy = task.meters[0].value["top_1"]
        self.assertAlmostEqual(accuracy, 1.0)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest
from test.generic.config_utils import get_fast_test_task_config

import torch
import torch.nn as nn
from classy_vision.models import ClassyModel
from classy_vision.models.classy_model_wrapper import ClassyModelWrapper
from classy_vision.tasks import build_task
from classy_vision.trainer import LocalTrainer
from torchvision import models


class TestModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 5)

    def forward(self, x):
        return self.linear(x)

    def extract_features(self, x):
        return torch.cat([x, x], dim=1)


class TestClassyModelWrapper(unittest.TestCase):
    def test_classy_model_wrapper(self):
        model = TestModel()
        classy_model = ClassyModelWrapper(model)
        # test that the returned object is an instance of ClassyModel
        self.assertIsInstance(classy_model, ClassyModel)

        # test that forward works correctly
        input = torch.zeros((100, 10))
        output = classy_model(input)
        self.assertEqual(output.shape, (100, 5))

        # test that extract_features works correctly
        input = torch.zeros((100, 10))
        output = classy_model.extract_features(input)
        self.assertEqual(output.shape, (100, 20))

        # test that get_classy_state and set_classy_state work
        nn.init.constant_(classy_model.model.linear.weight, 1)
        weights = copy.deepcopy(classy_model.model.linear.weight.data)
        state_dict = classy_model.get_classy_state(deep_copy=True)
        nn.init.constant_(classy_model.model.linear.weight, 0)
        classy_model.set_classy_state(state_dict)
        self.assertTrue(torch.allclose(weights, classy_model.model.linear.weight.data))

    def test_classy_model_wrapper_properties(self):
        # test that the properties work correctly when passed to the wrapper
        model = TestModel()
        num_classes = 5
        input_shape = (10,)
        output_shape = (num_classes,)
        model_depth = 1
        classy_model = ClassyModelWrapper(
            model,
            input_shape=input_shape,
            output_shape=output_shape,
            model_depth=model_depth,
        )
        self.assertEqual(classy_model.input_shape, input_shape)
        self.assertEqual(classy_model.output_shape, output_shape)
        self.assertEqual(classy_model.model_depth, model_depth)

    def test_train_step(self):
        # test that the model can be run in a train step
        model = models.resnet34(pretrained=False)
        classy_model = ClassyModelWrapper(model)

        config = get_fast_test_task_config()
        task = build_task(config)
        task.set_model(classy_model)
        trainer = LocalTrainer()
        trainer.train(task)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Optional

from classy_vision.generic.distributed_util import set_cpu_device, set_cuda_device_index

from .classy_trainer import ClassyTrainer


class LocalTrainer(ClassyTrainer):
    """Trainer to be used if you want want use only a single training process.
    """

    def __init__(
        self,
        use_gpu: Optional[bool] = None,
        num_dataloader_workers: int = 0,
        dataloader_mp_context: Optional[str] = None,
    ):
        """Constructor for LocalTrainer.

        Args:
            use_gpu: If true, then use GPU 0 for training.
                If None, then check if we have GPUs available, if we do
                then use GPU for training.
            num_dataloader_workers: Number of CPU processes doing dataloading
                per GPU. If 0, then dataloading is done on main thread.
            dataloader_mp_context: Determines how to launch
                new processes for dataloading. Must be one of "fork", "forkserver",
                "spawn". If None, process launching is inherited from parent.
        """
        super().__init__(
            use_gpu=use_gpu,
            num_dataloader_workers=num_dataloader_workers,
            dataloader_mp_context=dataloader_mp_context,
        )
        if self.use_gpu:
            logging.info("Using GPU, CUDA device index: {}".format(0))
            set_cuda_device_index(0)
        else:
            logging.info("Using CPU")
            set_cpu_device()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.nn as nn


class ClassyLoss(nn.Module):
    """
    Base class to calculate the loss during training.

    This implementation of :class:`torch.nn.Module` allows building
    the loss object from a configuration file.
    """

    def __init__(self):
        """
        Constructor for ClassyLoss.
        """
        super(ClassyLoss, self).__init__()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyLoss":
        """Instantiates a ClassyLoss from a configuration.

        Args:
            config: A configuration for a ClassyLoss.

        Returns:
            A ClassyLoss instance.
        """
        raise NotImplementedError()

    def forward(self, output, target):
        """
        Compute the loss for the provided sample.

        Refer to :class:`torch.nn.Module` for more details.
        """
        raise NotImplementedError



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import importlib
import logging
import os
import sys
from pathlib import Path


def import_all_modules(root, base_module):
    for file in os.listdir(root):
        if file.endswith((".py", ".pyc")) and not file.startswith("_"):
            module = file[: file.find(".py")]
            if module not in sys.modules:
                module_name = ".".join([base_module, module])
                importlib.import_module(module_name)


def import_all_packages_from_directory(root):
    """Automatically imports all packages under the root directory.

    For instance, if your directories look like:
        root / foo / __init__.py
        root / foo / abc.py
        root / bar.py
        root / baz / xyz.py

    This function will import the package foo, but not bar or baz."""

    for file in os.listdir(root):
        # Try to import each file in the directory. Our previous implementation
        # would look for directories here and see if there's a __init__.py
        # under that directory, but that turns out to be unreliable while
        # running on AWS: EFS filesystems cache metadata bits so the directory
        # and existence checks fail even when the import succeeds. We should
        # find a better workaround eventually, but this will do for now.
        try:
            file = Path(file)
            module_name = file.name
            # Dots have special meaning in Python packages -- it's a relative
            # import or a subpackage. Skip these.
            if "." not in module_name and module_name not in sys.modules:
                logging.debug(f"Automatically importing {module_name}")
                importlib.import_module(module_name)
        except ModuleNotFoundError:
            pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import unittest
from typing import Any, Callable, Dict, Optional, Union

import torch.nn as nn
import torch.nn.functional as F
from classy_vision.dataset import ClassyDataset, register_dataset
from classy_vision.dataset.core.random_image_datasets import (
    RandomImageBinaryClassDataset,
    SampleType,
)
from classy_vision.dataset.transforms import (
    ClassyTransform,
    GenericImageTransform,
    build_transforms,
)
from classy_vision.losses import ClassyLoss, register_loss
from classy_vision.models import ClassyModel, register_model
from classy_vision.optim import SGD
from classy_vision.optim.param_scheduler import ConstantParamScheduler
from classy_vision.tasks import ClassificationTask
from classy_vision.trainer import LocalTrainer
from torchvision import transforms


# WARNING: The goal of this test is to use our public API as advertised in our
# tutorials and make sure everything trains successfully. If you break this
# test, make sure you also update our tutorials.


@register_dataset("my_dataset")
class MyDataset(ClassyDataset):
    def __init__(
        self,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: int,
        crop_size: int,
        class_ratio: float,
        seed: int,
        split: Optional[str] = None,
    ) -> None:
        dataset = RandomImageBinaryClassDataset(
            crop_size, class_ratio, num_samples, seed, SampleType.TUPLE
        )
        super().__init__(
            dataset, split, batchsize_per_replica, shuffle, transform, num_samples
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "MyDataset":
        assert all(key in config for key in ["crop_size", "class_ratio", "seed"])

        split = config.get("split")
        crop_size = config["crop_size"]
        class_ratio = config["class_ratio"]
        seed = config["seed"]
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = cls.parse_config(config)
        transform = build_transforms(transform_config)
        return cls(
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            crop_size,
            class_ratio,
            seed,
            split=split,
        )


@register_loss("my_loss")
class MyLoss(ClassyLoss):
    def forward(self, input, target):
        labels = F.one_hot(target, num_classes=2).float()
        return F.binary_cross_entropy(input, labels)

    @classmethod
    def from_config(cls, config):
        # We don't need anything from the config
        return cls()


@register_model("my_model")
class MyModel(ClassyModel):
    def __init__(self):
        super().__init__()
        self.model = nn.Sequential(
            nn.AdaptiveAvgPool2d((20, 20)),
            nn.Flatten(1),
            nn.Linear(3 * 20 * 20, 2),
            nn.Sigmoid(),
        )

    def forward(self, x):
        x = self.model(x)
        return x

    @classmethod
    def from_config(cls, config):
        return cls()


class APITest(unittest.TestCase):
    def testOne(self):
        train_dataset = MyDataset(
            batchsize_per_replica=32,
            shuffle=False,
            transform=GenericImageTransform(
                transform=transforms.Compose(
                    [
                        transforms.RandomResizedCrop(224),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToTensor(),
                        transforms.Normalize(
                            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                        ),
                    ]
                )
            ),
            num_samples=100,
            crop_size=224,
            class_ratio=0.5,
            seed=0,
        )

        test_dataset = MyDataset(
            batchsize_per_replica=32,
            shuffle=False,
            transform=GenericImageTransform(
                transform=transforms.Compose(
                    [
                        transforms.Resize(256),
                        transforms.CenterCrop(224),
                        transforms.ToTensor(),
                        transforms.Normalize(
                            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                        ),
                    ]
                )
            ),
            num_samples=100,
            crop_size=224,
            class_ratio=0.5,
            seed=0,
        )

        model = MyModel()
        loss = MyLoss()

        optimizer = SGD(lr_scheduler=ConstantParamScheduler(0.01))

        task = (
            ClassificationTask()
            .set_model(model)
            .set_dataset(train_dataset, "train")
            .set_dataset(test_dataset, "test")
            .set_loss(loss)
            .set_optimizer(optimizer)
            .set_num_epochs(1)
        )

        trainer = LocalTrainer()
        trainer.train(task)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
from typing import Any, Callable, Dict, List, Optional, Union

import torch
import torchvision.transforms as transforms
import torchvision.transforms._transforms_video as transforms_video

from . import ClassyTransform, build_transforms, register_transform
from .util import ApplyTransformToKey, ImagenetConstants


class VideoConstants:
    """Constant variables related to the video classification.

    Use the same mean/std from image classification to enable the parameter
    inflation where parameters of 2D conv in image model can be inflated into
    3D conv in video model.

    MEAN: often used to be subtracted from pixel RGB value.
    STD: often used to divide the pixel RGB value after mean centering.
    SIZE_RANGE: a (min_size, max_size) tuple which denotes the range of
        size of the rescaled video clip.
    CROP_SIZE: the size of spatial cropping in the video clip.
    """

    MEAN = ImagenetConstants.MEAN  #
    STD = ImagenetConstants.STD
    SIZE_RANGE = (128, 160)
    CROP_SIZE = 112


def _get_rescaled_size(scale, h, w):
    if h < w:
        new_h = scale
        new_w = int(scale * w / h)
    else:
        new_w = scale
        new_h = int(scale * h / w)
    return new_h, new_w


@register_transform("video_clip_random_resize_crop")
class VideoClipRandomResizeCrop(ClassyTransform):
    """A video clip transform that is often useful for trainig data.

    Given a size range, randomly choose a size. Rescale the clip so that
    its short edge equals to the chosen size. Then randomly crop the video
    clip with the specified size.
    Such training data augmentation is used in VGG net
    (https://arxiv.org/abs/1409.1556).
    Also see reference implementation `Kinetics.spatial_sampling` in SlowFast
        codebase.
    """

    def __init__(
        self,
        crop_size: Union[int, List[int]],
        size_range: List[int],
        interpolation_mode: str = "bilinear",
    ):
        """The constructor method of VideoClipRandomResizeCrop class.

        Args:
            crop_size: int or 2-tuple as the expected output crop_size (height, width)
            size_range: the min- and max size
            interpolation_mode: Default: "bilinear"

        """
        if isinstance(crop_size, tuple):
            assert len(crop_size) == 2, "crop_size should be tuple (height, width)"
            self.crop_size = crop_size
        else:
            self.crop_size = (crop_size, crop_size)

        self.interpolation_mode = interpolation_mode
        self.size_range = size_range

    def __call__(self, clip):
        """Callable function which applies the tranform to the input clip.

        Args:
            clip (torch.Tensor): input clip tensor

        """
        # clip size: C x T x H x W
        rand_size = random.randint(self.size_range[0], self.size_range[1])
        new_h, new_w = _get_rescaled_size(rand_size, clip.size()[2], clip.size()[3])
        clip = torch.nn.functional.interpolate(
            clip, size=(new_h, new_w), mode=self.interpolation_mode
        )
        assert (
            self.crop_size[0] <= new_h and self.crop_size[1] <= new_w
        ), "crop size can not be larger than video frame size"

        i = random.randint(0, new_h - self.crop_size[0])
        j = random.randint(0, new_w - self.crop_size[1])
        clip = clip[:, :, i : i + self.crop_size[0], j : j + self.crop_size[1]]
        return clip


@register_transform("video_clip_resize")
class VideoClipResize(ClassyTransform):
    """A video clip transform that is often useful for testing data.

    Given an input size, rescale the clip so that its short edge equals to
    the input size while aspect ratio is preserved.
    """

    def __init__(self, size: int, interpolation_mode: str = "bilinear"):
        """The constructor method of VideoClipResize class.

        Args:
            size: input size
            interpolation_mode: Default: "bilinear". See valid values in
                (https://pytorch.org/docs/stable/nn.functional.html#torch.nn.
                functional.interpolate)

        """
        self.interpolation_mode = interpolation_mode
        self.size = size

    def __call__(self, clip):
        """Callable function which applies the tranform to the input clip.

        Args:
            clip (torch.Tensor): input clip tensor

        """
        # clip size: C x T x H x W
        if not min(clip.size()[2], clip.size()[3]) == self.size:
            new_h, new_w = _get_rescaled_size(self.size, clip.size()[2], clip.size()[3])
            clip = torch.nn.functional.interpolate(
                clip, size=(new_h, new_w), mode=self.interpolation_mode
            )
        return clip


@register_transform("video_tuple_to_map_transform")
class VideoTupleToMapTransform(ClassyTransform):
    """A video transform which maps video data from tuple to dict.

    It takes a sample of the form (video, audio, target) and returns a sample of
    the form {"input": {"video" video, "audio": audio}, "target": target}. If
    the sample is a map with these keys already present, it will pass the sample
    through.

    It's particularly useful for remapping torchvision samples which are
    tuples of the form (video, audio, target).
    """

    def __call__(self, sample):
        """Callable function which applies the tranform to the input sample data.

        Args:
            sample: input sample data that will undergo the transform

        """
        # If sample is a map and already has input / target keys, pass through
        if isinstance(sample, dict):
            assert "input" in sample and "target" in sample, (
                "Input to tuple to map transform must be a tuple of length 3 "
                "or a dict with keys 'input' and 'target'"
            )
            assert (
                "video" in sample["input"] and "audio" in sample["input"]
            ), "Input data must include video / audio fields"
            return sample

        # Should be a tuple (or other sequential) of length 3, transform to map
        assert len(sample) == 3, "Sequential must be length 3 for conversion"
        video, audio, target = sample
        output_sample = {"input": {"video": video, "audio": audio}, "target": target}
        return output_sample


@register_transform("video_default_augment")
class VideoDefaultAugmentTransform(ClassyTransform):
    """This is the default video transform with data augmentation which is useful for
    training.

    It sequentially prepares a torch.Tensor of video data, randomly
    resizes the video clip, takes a random spatial cropping, randomly flips the
    video clip horizontally, and normalizes the pixel values by mean subtraction
    and standard deviation division.

    """

    def __init__(
        self,
        crop_size: Union[int, List[int]] = VideoConstants.CROP_SIZE,
        size_range: List[int] = VideoConstants.SIZE_RANGE,
        mean: List[float] = VideoConstants.MEAN,
        std: List[float] = VideoConstants.STD,
    ):
        """The constructor method of VideoDefaultAugmentTransform class.

        Args:
            crop_size: expected output crop_size (height, width)
            size_range : a 2-tuple denoting the min- and max size
            mean: a 3-tuple denoting the pixel RGB mean
            std: a 3-tuple denoting the pixel RGB standard deviation

        """

        self._transform = transforms.Compose(
            [
                transforms_video.ToTensorVideo(),
                # TODO(zyan3): migrate VideoClipRandomResizeCrop to TorchVision
                VideoClipRandomResizeCrop(crop_size, size_range),
                transforms_video.RandomHorizontalFlipVideo(),
                transforms_video.NormalizeVideo(mean=mean, std=std),
            ]
        )

    def __call__(self, video):
        """Apply the default transform with data augmentation to video.

        Args:
            video: input video that will undergo the transform

        """
        return self._transform(video)


@register_transform("video_default_no_augment")
class VideoDefaultNoAugmentTransform(ClassyTransform):
    """This is the default video transform without data augmentation which is useful
    for testing.

    It sequentially prepares a torch.Tensor of video data, resize the
    video clip to have the specified short edge, and normalize the pixel values
    by mean subtraction and standard deviation division.

    """

    def __init__(
        self,
        size: int = VideoConstants.SIZE_RANGE[0],
        mean: List[float] = VideoConstants.MEAN,
        std: List[float] = VideoConstants.STD,
    ):
        """The constructor method of VideoDefaultNoAugmentTransform class.

        Args:
            size: the short edge of rescaled video clip
            mean: a 3-tuple denoting the pixel RGB mean
            std: a 3-tuple denoting the pixel RGB standard deviation

        """
        self._transform = transforms.Compose(
            # At testing stage, central cropping is not used because we
            # conduct fully convolutional-style testing
            [
                transforms_video.ToTensorVideo(),
                # TODO(zyan3): migrate VideoClipResize to TorchVision
                VideoClipResize(size),
                transforms_video.NormalizeVideo(mean=mean, std=std),
            ]
        )

    def __call__(self, video):
        """Apply the default transform without data augmentation to video.

        Args:
            video: input video that will undergo the transform

        """
        return self._transform(video)


@register_transform("dummy_audio_transform")
class DummyAudioTransform(ClassyTransform):
    """This is a dummy audio transform.

    It ignores actual audio data, and returns an empty tensor. It is useful when
    actual audio data is raw waveform and has a varying number of waveform samples
    which makes minibatch assembling impossible

    """

    def __init__(self):
        """The constructor method of DummyAudioTransform class.
        """

        pass

    def __call__(self, _audio):
        """Callable function which applies the tranform to the input audio data.

        Args:
            audio: input audio data that will undergo the dummy transform

        """
        return torch.zeros(0, 1, dtype=torch.float)


class ClassyVideoGenericTransform(object):
    """This is a generic video transform which includes both video transform
    and audio transform.
    """

    def __init__(
        self,
        config: Optional[Dict[str, List[Dict[str, Any]]]] = None,
        split: str = "train",
    ):
        """The constructor method of ClassyVideoGenericTransform class.

        Args:
            config: If provided, it is a dict where key is the data modality, and
                value is a dict specifying the transform config
            split: the split of the data to which the transform will be applied
        """
        self.transforms = {
            "video": VideoDefaultAugmentTransform()
            if split == "train"
            else VideoDefaultNoAugmentTransform(),
            "audio": DummyAudioTransform(),
        }
        if config is not None:
            for mode, modal_config in config.items():
                assert mode in ["video", "audio"], (
                    "unknown video data modality %s" % mode
                )
                self.transforms[mode] = build_transforms(modal_config)

    def __call__(self, video: Dict):
        """Callable function which applies the tranform to the input video data.

        Args:
            video: input video data that will undergo the transform

        """
        assert isinstance(video, dict), "video data is expected be a dict"
        for mode, modal_data in video.items():
            if mode in self.transforms:
                video[mode] = self.transforms[mode](modal_data)
        return video


DEFAULT_KEY_MAP = VideoTupleToMapTransform()


def build_video_field_transform_default(
    config: Optional[Dict[str, List[Dict[str, Any]]]],
    split: str = "train",
    key: str = "input",
    key_map_transform: Optional[Callable] = DEFAULT_KEY_MAP,
) -> Callable:
    """Returns transform that first maps sample to video keys, then
    returns a transform on the specified key in dict.

    Converts tuple (list, etc) sample to dict with input / target keys.
    For a dict sample, verifies that dict has input / target keys.
    For all other samples throws.

    Args:
        config: If provided, it is a dict where key is the data modality, and
            value is a dict specifying the transform config
        split: the split of the data to which the transform will be applied
        key: the key in data sample of type dict whose corresponding value will
            undergo the transform
        key_map_transform: If provided, it is a transform which maps sample of type
            tuple to sample of type dict. See default value VideoTupleToMapTransform()
            as an example

    """
    transform = ApplyTransformToKey(ClassyVideoGenericTransform(config, split), key=key)
    if key_map_transform is None:
        return transform

    return transforms.Compose([key_map_transform, transform])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules

from .classy_meter import ClassyMeter


FILE_ROOT = Path(__file__).parent


METER_REGISTRY = {}


def build_meter(config):
    """Builds a ClassyMeter from a config.

    This assumes a 'name' key in the config which is used to determine what
    meter class to instantiate. For instance, a config `{"name": "my_meter",
    "foo": "bar"}` will find a class that was registered as "my_meter" (see
    :func:`register_meter`) and call .from_config on it."""
    return METER_REGISTRY[config["name"]].from_config(config)


def build_meters(config):
    configs = [{"name": name, **args} for name, args in config.items()]
    return [build_meter(config) for config in configs]


def register_meter(name):
    """Registers a ClassyMeter subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyMeter from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyMeter subclass, like this:

        @register_meter('accuracy')
        class AccuracyMeter(ClassyMeter):
            ...

    To instantiate a meter from a configuration file, see
    :func:`build_meter`."""

    def register_meter_cls(cls):
        if name in METER_REGISTRY:
            raise ValueError("Cannot register duplicate meter ({})".format(name))
        if not issubclass(cls, ClassyMeter):
            raise ValueError(
                "Meter ({}: {}) must extend \
                ClassyMeter".format(
                    name, cls.__name__
                )
            )
        METER_REGISTRY[name] = cls
        return cls

    return register_meter_cls


# automatically import any Python files in the meters/ directory
import_all_modules(FILE_ROOT, "classy_vision.meters")

from .accuracy_meter import AccuracyMeter  # isort:skip
from .precision_meter import PrecisionAtKMeter  # isort:skip
from .recall_meter import RecallAtKMeter  # isort:skip
from .video_accuracy_meter import VideoAccuracyMeter  # isort:skip

__all__ = [
    "AccuracyMeter",
    "ClassyMeter",
    "PrecisionAtKMeter",
    "RecallAtKMeter",
    "VideoAccuracyMeter",
    "build_meter",
    "build_meters",
    "register_meter",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.dataset import build_dataset, register_dataset
from classy_vision.dataset.classy_video_dataset import (
    ClassyVideoDataset,
    MaxLengthClipSampler,
)
from classy_vision.dataset.core import ListDataset
from classy_vision.dataset.transforms.util_video import (
    build_video_field_transform_default,
)
from torch.utils.data import Sampler


DUMMY_SAMPLES_1 = [
    {
        "input": {
            "video": torch.randint(0, 256, (8, 3, 128, 128), dtype=torch.uint8),
            "audio": torch.rand(1000, 1, dtype=torch.float32),
        },
        "target": torch.tensor([[0]]),
    }
]


DUMMY_CONFIG = {
    "name": "test_video_dataset",
    "split": "train",
    "batchsize_per_replica": 1,
    "use_shuffle": True,
    "num_samples": 1,
    "frames_per_clip": 8,
    "video_dir": "dummy_video_dir",
}


class MockClipSampler(Sampler):
    def __init__(self, full_size=1000):
        self.full_size = full_size

    def __iter__(self):
        indices = list(range(self.full_size))
        return iter(indices)

    def __len__(self):
        return self.full_size


@register_dataset("test_video_dataset")
class TestVideoDataset(ClassyVideoDataset):
    """Test dataset for validating registry functions"""

    def __init__(
        self,
        split,
        batchsize_per_replica,
        shuffle,
        transform,
        num_samples,
        clips_per_video,
        samples,
    ):
        self.samples = samples
        input_tensors = [sample["input"] for sample in samples]
        target_tensors = [sample["target"] for sample in samples]
        dataset = ListDataset(input_tensors, target_tensors, loader=lambda x: x)
        super(TestVideoDataset, self).__init__(
            dataset,
            split,
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            clips_per_video,
        )

    @classmethod
    def from_config(cls, config, samples):
        split = config.get("split")
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            step_between_clips,
            frame_rate,
            clips_per_video,
        ) = cls.parse_config(config)
        transform = build_video_field_transform_default(transform_config, split)
        return cls(
            split,
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            clips_per_video,
            samples,
        )


class TestRegistryFunctions(unittest.TestCase):
    """
    Tests functions that use registry
    """

    def test_build_dataset(self):
        dataset = build_dataset(DUMMY_CONFIG, DUMMY_SAMPLES_1)
        self.assertTrue(isinstance(dataset, TestVideoDataset))


class TestClassyVideoDataset(unittest.TestCase):
    """
    Tests member functions of ClassyVideoDataset.
    """

    def setUp(self):
        self.dataset = build_dataset(DUMMY_CONFIG, DUMMY_SAMPLES_1)

    def test_parse_config(self):
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            step_between_clips,
            frame_rate,
            clips_per_video,
        ) = self.dataset.parse_config(DUMMY_CONFIG)

    def test_max_length_clip_sampler(self):
        clip_sampler = MockClipSampler(full_size=1000)
        clip_sampler = MaxLengthClipSampler(clip_sampler, num_samples=64)
        count = 0
        for _clip_index in iter(clip_sampler):
            count += 1
        self.assertEqual(count, 64)
        self.assertEqual(len(clip_sampler), 64)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.cosine_scheduler import CosineParamScheduler


class TestCosineScheduler(unittest.TestCase):
    _num_epochs = 10

    def _get_valid_decay_config(self):
        return {"name": "cosine", "start_lr": 0.1, "end_lr": 0}

    def _get_valid_decay_config_intermediate_values(self):
        return [0.0976, 0.0905, 0.0794, 0.0655, 0.05, 0.0345, 0.0206, 0.0095, 0.0024]

    def test_invalid_config(self):
        # Invalid num epochs
        config = self._get_valid_decay_config()

        bad_config = copy.deepcopy(config)
        # Invalid Base lr
        del bad_config["start_lr"]
        with self.assertRaises(AssertionError):
            CosineParamScheduler.from_config(bad_config)

        # Invalid end_lr
        bad_config["start_lr"] = config["start_lr"]
        del bad_config["end_lr"]
        with self.assertRaises(AssertionError):
            CosineParamScheduler.from_config(bad_config)

    def test_scheduler_as_decay(self):
        config = self._get_valid_decay_config()

        scheduler = CosineParamScheduler.from_config(config)
        schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            config["start_lr"]
        ] + self._get_valid_decay_config_intermediate_values()

        self.assertEqual(schedule, expected_schedule)

    def test_scheduler_as_warmup(self):
        config = self._get_valid_decay_config()
        # Swap start and end lr to change to warmup
        tmp = config["start_lr"]
        config["start_lr"] = config["end_lr"]
        config["end_lr"] = tmp

        scheduler = CosineParamScheduler.from_config(config)
        schedule = [
            round(scheduler(epoch_num / self._num_epochs), 4)
            for epoch_num in range(self._num_epochs)
        ]
        # Schedule should be decay reversed
        expected_schedule = [config["start_lr"]] + list(
            reversed(self._get_valid_decay_config_intermediate_values())
        )

        self.assertEqual(schedule, expected_schedule)

    def test_scheduler_warmup_decay_match(self):
        decay_config = self._get_valid_decay_config()
        decay_scheduler = CosineParamScheduler.from_config(decay_config)

        warmup_config = copy.deepcopy(decay_config)
        # Swap start and end lr to change to warmup
        tmp = warmup_config["start_lr"]
        warmup_config["start_lr"] = warmup_config["end_lr"]
        warmup_config["end_lr"] = tmp
        warmup_scheduler = CosineParamScheduler.from_config(warmup_config)

        decay_schedule = [
            round(decay_scheduler(epoch_num / 1000), 8) for epoch_num in range(1, 1000)
        ]
        warmup_schedule = [
            round(warmup_scheduler(epoch_num / 1000), 8) for epoch_num in range(1, 1000)
        ]

        self.assertEqual(decay_schedule, list(reversed(warmup_schedule)))

    def test_build_cosine_scheduler(self):
        config = self._get_valid_decay_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, CosineParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest.mock import Mock

from classy_vision.dataset import build_dataset
from classy_vision.hooks import ClassyHook
from classy_vision.losses import build_loss
from classy_vision.models import build_model
from classy_vision.optim import build_optimizer
from classy_vision.optim.param_scheduler import UpdateInterval
from classy_vision.tasks import ClassificationTask, ClassyTask
from classy_vision.trainer import LocalTrainer


class TestParamSchedulerIntegration(unittest.TestCase):
    def _get_config(self):
        return {
            "loss": {"name": "CrossEntropyLoss"},
            "dataset": {
                "train": {
                    "name": "synthetic_image",
                    "split": "train",
                    "num_classes": 2,
                    "crop_size": 20,
                    "class_ratio": 0.5,
                    "num_samples": 10,
                    "seed": 0,
                    "batchsize_per_replica": 5,
                    "use_shuffle": True,
                    "transforms": [
                        {
                            "name": "apply_transform_to_key",
                            "transforms": [
                                {"name": "ToTensor"},
                                {
                                    "name": "Normalize",
                                    "mean": [0.485, 0.456, 0.406],
                                    "std": [0.229, 0.224, 0.225],
                                },
                            ],
                            "key": "input",
                        }
                    ],
                },
                "test": {
                    "name": "synthetic_image",
                    "split": "test",
                    "num_classes": 2,
                    "crop_size": 20,
                    "class_ratio": 0.5,
                    "num_samples": 10,
                    "seed": 0,
                    "batchsize_per_replica": 5,
                    "use_shuffle": False,
                    "transforms": [
                        {
                            "name": "apply_transform_to_key",
                            "transforms": [
                                {"name": "ToTensor"},
                                {
                                    "name": "Normalize",
                                    "mean": [0.485, 0.456, 0.406],
                                    "std": [0.229, 0.224, 0.225],
                                },
                            ],
                            "key": "input",
                        }
                    ],
                },
            },
            "model": {
                "name": "mlp",
                # 3x20x20 = 1200
                "input_dim": 1200,
                "output_dim": 1000,
                "hidden_dims": [10],
            },
            "meters": {"accuracy": {"topk": [1]}},
            "optimizer": {
                "name": "sgd",
                "num_epochs": 10,
                "lr": 0.1,
                "weight_decay": 1e-4,
                "momentum": 0.9,
            },
        }

    def _build_task(self, num_epochs):
        config = self._get_config()
        config["optimizer"]["num_epochs"] = num_epochs
        task = (
            ClassificationTask()
            .set_num_epochs(num_epochs)
            .set_loss(build_loss(config["loss"]))
            .set_model(build_model(config["model"]))
            .set_optimizer(build_optimizer(config["optimizer"]))
        )
        for phase_type in ["train", "test"]:
            dataset = build_dataset(config["dataset"][phase_type])
            task.set_dataset(dataset, phase_type)

        self.assertTrue(task is not None)
        return task

    def test_param_scheduler_epoch(self):
        task = self._build_task(num_epochs=3)

        where_list = []

        def scheduler_mock(where):
            where_list.append(where)
            return 0.1

        mock = Mock(side_effect=scheduler_mock)
        mock.update_interval = UpdateInterval.EPOCH
        task.optimizer.lr_scheduler = mock

        trainer = LocalTrainer()
        trainer.train(task)

        self.assertEqual(where_list, [0, 1 / 3, 2 / 3])

    def test_param_scheduler_step(self):
        task = self._build_task(num_epochs=3)

        where_list = []

        def scheduler_mock(where):
            where_list.append(where)
            return 0.1

        mock = Mock(side_effect=scheduler_mock)
        mock.update_interval = UpdateInterval.STEP
        task.optimizer.lr_scheduler = mock

        trainer = LocalTrainer()
        trainer.train(task)

        # We have 10 samples, batch size is 5. Each epoch is done in two steps.
        self.assertEqual(where_list, [0, 1 / 6, 2 / 6, 3 / 6, 4 / 6, 5 / 6])

    def test_hook(self):
        task = self._build_task(num_epochs=3)

        lr_list = []

        class TestHook(ClassyHook):
            on_rendezvous = ClassyHook._noop
            on_start = ClassyHook._noop
            on_phase_start = ClassyHook._noop
            on_sample = ClassyHook._noop
            on_forward = ClassyHook._noop
            on_loss_and_meter = ClassyHook._noop
            on_backward = ClassyHook._noop
            on_phase_end = ClassyHook._noop
            on_end = ClassyHook._noop

            def on_update(self, task: ClassyTask, local_variables) -> None:
                lr_list.append(task.optimizer.lr)

        task.set_hooks([TestHook()])

        def scheduler_mock(where):
            return where

        mock = Mock(side_effect=scheduler_mock)
        mock.update_interval = UpdateInterval.STEP
        task.optimizer.lr_scheduler = mock

        trainer = LocalTrainer()
        trainer.train(task)

        # We have 10 samples, batch size is 5. Each epoch is done in two steps.
        self.assertEqual(lr_list, [0, 1 / 6, 2 / 6, 3 / 6, 4 / 6, 5 / 6])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import collections
import logging
from typing import Any, Dict

from classy_vision import tasks
from classy_vision.generic.distributed_util import is_master
from classy_vision.generic.util import flatten_dict
from classy_vision.generic.visualize import plot_learning_curves
from classy_vision.hooks.classy_hook import ClassyHook


try:
    from visdom import Visdom

    visdom_available = True
except ImportError:
    visdom_available = False


class VisdomHook(ClassyHook):
    """Plots metrics on to Visdom.

    Visdom is a flexible tool for creating, organizing, and sharing visualizations
        of live, rich data. It supports Python.

    See more details at https://github.com/facebookresearch/visdom.

    """

    on_rendezvous = ClassyHook._noop
    on_start = ClassyHook._noop
    on_phase_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_update = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(
        self, server: str, port: str, env: str = "main", title_suffix: str = ""
    ) -> None:
        """
        Args:
            server: host name of the visdom server
            port: port of visdom server, such as 8097
            env: environment of visdom
            title_suffix: suffix that will be appended to the title
        """
        super().__init__()
        if not visdom_available:
            raise RuntimeError("Visdom is not installed, cannot use VisdomHook")

        self.server: str = server
        self.port: str = port
        self.env: str = env
        self.title_suffix: str = title_suffix

        self.metrics: Dict = {}
        self.visdom: Visdom = Visdom(self.server, self.port)

    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Plot the metrics on visdom.
        """
        phase_type = task.phase_type
        metrics = self.metrics
        batches = len(task.losses)

        if batches == 0:
            return

        # Loss for the phase
        loss = sum(task.losses) / (batches * task.get_batchsize_per_replica())
        loss_key = phase_type + "_loss"
        if loss_key not in metrics:
            metrics[loss_key] = []
        metrics[loss_key].append(loss)

        # Optimizer LR for the phase
        optimizer_lr = task.optimizer.lr
        lr_key = phase_type + "_learning_rate"
        if lr_key not in metrics:
            metrics[lr_key] = []
        metrics[lr_key].append(optimizer_lr)

        # Calculate meters
        for meter in task.meters:
            if isinstance(meter.value, collections.MutableMapping):
                flattened_meters_dict = flatten_dict(meter.value, prefix=meter.name)
                for k, v in flattened_meters_dict.items():
                    metric_key = phase_type + "_" + k
                    if metric_key not in metrics:
                        metrics[metric_key] = []
                    metrics[metric_key].append(v)
            else:
                metric_key = phase_type + "_" + meter.name
                if metric_key not in metrics:
                    metrics[metric_key] = []
                metrics[metric_key].append(meter.value)

        # update learning curve visualizations:
        phase_type = "train" if task.train else "test"
        title = "%s-%s-%d" % (
            phase_type,
            task.base_model.__class__.__name__,
            task.base_model.model_depth,
        )
        title += self.title_suffix

        if not task.train and is_master():
            logging.info("Plotting learning curves to visdom")
            plot_learning_curves(
                metrics, visdom_server=self.visdom, env=self.env, win=title, title=title
            )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections.abc import Sequence
from typing import Any, Dict, List, Optional

import torch.nn as nn
from classy_vision.generic.util import is_pos_int
from classy_vision.heads import ClassyHead, register_head


class FullyConvolutionalLinear(nn.Module):
    def __init__(self, dim_in, num_classes, act_func="softmax"):
        super(FullyConvolutionalLinear, self).__init__()
        # Perform FC in a fully convolutional manner. The FC layer will be
        # initialized with a different std comparing to convolutional layers.
        self.projection = nn.Linear(dim_in, num_classes, bias=True)

        # Softmax for evaluation and testing.
        if act_func == "softmax":
            self.act = nn.Softmax(dim=4)
        elif act_func == "sigmoid":
            self.act = nn.Sigmoid()
        else:
            raise NotImplementedError(
                "{} is not supported as an activation" "function.".format(act_func)
            )

    def forward(self, x):
        # (N, C, T, H, W) -> (N, T, H, W, C).
        x = x.permute((0, 2, 3, 4, 1))
        x = self.projection(x)
        # Performs fully convlutional inference.
        if not self.training:
            x = self.act(x)
            x = x.mean([1, 2, 3])
        x = x.view(x.shape[0], -1)
        return x


@register_head("fully_convolutional_linear")
class FullyConvolutionalLinearHead(ClassyHead):
    """
    This head defines a 3d average pooling layer (:class:`torch.nn.AvgPool3d`)
    followed by a fully convolutional linear layer. This layer performs a
    fully-connected projection during training, when the input size is 1x1x1.
    It performs a convolutional projection during testing when the input size
    is larger than 1x1x1.
    """

    def __init__(
        self,
        unique_id: str,
        num_classes: int,
        in_plane: int,
        pool_size: List[int],
        activation_func: str,
        use_dropout: Optional[bool] = None,
    ):
        """
        Constructor for FullyConvolutionalLinearHead.

        Args:
            unique_id: A unique identifier for the head. Multiple instances of
                the same head might be attached to a model, and unique_id is used
                to refer to them.
            num_classes: Number of classes for the head.
            in_plane: Input size for the fully connected layer.
            pool_size: Kernel size for the 3d pooling layer.
            activation_func: activation function to use. 'softmax': applies
                softmax on the output. 'sigmoid': applies sigmoid on the output.
            use_dropout: Whether to apply dropout after the pooling layer.
        """
        super().__init__(unique_id, num_classes)
        self.final_avgpool = nn.AvgPool3d(pool_size, stride=1)
        if use_dropout:
            self.dropout = nn.Dropout(p=0.5)
        # we separate average pooling from the fully-convolutional linear projection
        # because for multi-path models such as SlowFast model, the input can be
        # more than 1 tesnor. In such case, we can define a new head to combine multiple
        # tensors via concat or addition, do average pooling, but still reuse
        # FullyConvolutionalLinear inside of it.
        self.head_fcl = FullyConvolutionalLinear(
            in_plane, num_classes, act_func=activation_func
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "FullyConvolutionalLinearHead":
        """Instantiates a FullyConvolutionalLinearHead from a configuration.

        Args:
            config: A configuration for a FullyConvolutionalLinearHead.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A FullyConvolutionalLinearHead instance.
        """
        required_args = ["pool_size", "in_plane", "num_classes"]
        for arg in required_args:
            assert arg in config, "argument %s is required" % arg

        config.update({"activation_func": config.get("activation_func", "softmax")})
        config.update({"use_dropout": config.get("use_dropout", False)})

        assert (
            isinstance(config["pool_size"], Sequence) and len(config["pool_size"]) == 3
        )
        for pool_size_dim in config["pool_size"]:
            assert is_pos_int(pool_size_dim)
        assert is_pos_int(config["in_plane"])
        assert is_pos_int(config["num_classes"])

        num_classes = config.get("num_classes", None)
        in_plane = config["in_plane"]
        return cls(
            config["unique_id"],
            num_classes,
            in_plane,
            config["pool_size"],
            config["activation_func"],
            config["use_dropout"],
        )

    def forward(self, x):
        out = self.final_avgpool(x)
        if hasattr(self, "dropout"):
            out = self.dropout(out)
        out = self.head_fcl(out)
        return out



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
from enum import Enum
from typing import Any, Dict

import torch
import torch.nn as nn
from classy_vision.heads.classy_head import ClassyHead

from .classy_block import ClassyBlock


class ClassyModelEvaluationMode(Enum):
    DEFAULT = 0
    VIDEO_CLIP_AVERAGING = 1


class ClassyModel(nn.Module):
    """Base class for models in classy vision.

    A model refers either to a specific architecture (e.g. ResNet50) or a
    family of architectures (e.g. ResNet). Models can take arguments in the
    constructor in order to configure different behavior (e.g.
    hyperparameters).  Classy Models must implement :method:`from_config` in
    order to allow instantiation from a configuration file. Like regular
    PyTorch models, Classy Models must also implement :method:`forward`, where
    the bulk of the inference logic lives.

    Classy Models also have some advanced functionality for production
    fine-tuning systems. For example, we allow users to train a trunk
    model and then attach heads to the model via the attachable
    blocks.  Making your model support the trunk-heads paradigm is
    completely optional.

    """

    def __init__(self):
        """Constructor for ClassyModel."""
        super().__init__()

        self._attachable_blocks = {}
        self._heads = nn.ModuleDict()
        self._head_outputs = {}

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyModel":
        """Instantiates a ClassyModel from a configuration.

        Args:
            config: A configuration for the ClassyModel.

        Returns:
            A ClassyModel instance.
        """
        raise NotImplementedError

    @classmethod
    def from_checkpoint(cls, checkpoint):
        from . import build_model

        model = build_model(checkpoint["input_args"]["config"]["model"])
        model.set_classy_state(checkpoint["classy_state_dict"]["base_model"])
        return model

    def get_classy_state(self, deep_copy=False):
        """Get the state of the ClassyModel.

        The returned state is used for checkpointing.

        Args:
            deep_copy: If True, creates a deep copy of the state Dict. Otherwise, the
                returned Dict's state will be tied to the object's.

        Returns:
            A state dictionary containing the state of the model.
        """
        # If the model doesn't have head for fine-tuning, all of model's state
        # live in the trunk
        attached_heads = self.get_heads()
        # clear heads to get trunk only states. There shouldn't be any component
        # states depend on heads
        self._clear_heads()
        trunk_state_dict = super().state_dict()
        self.set_heads(attached_heads)

        head_state_dict = {}
        for block, heads in attached_heads.items():
            head_state_dict[block] = {
                head_name: head.state_dict() for head_name, head in heads.items()
            }
        model_state_dict = {
            "model": {"trunk": trunk_state_dict, "heads": head_state_dict}
        }
        if deep_copy:
            model_state_dict = copy.deepcopy(model_state_dict)
        return model_state_dict

    def load_head_states(self, state):
        """Load only the state (weights) of the heads.

        For a trunk-heads model, this function allows the user to
        only update the head state of the model. Useful for attaching
        fine-tuned heads to a pre-trained trunk.

        Args:
            state (Dict): Contains the classy model state under key "model"

        """
        for block_name, head_states in state["model"]["heads"].items():
            for head_name, head_state in head_states.items():
                self._heads[block_name][head_name].load_state_dict(head_state)

    def set_classy_state(self, state):
        """Set the state of the ClassyModel.

        Args:
            state_dict: The state dictionary. Must be the output of a call to
                :method:`get_classy_state`.

        This is used to load the state of the model from a checkpoint.
        """
        self.load_head_states(state)

        current_state = self.state_dict()
        current_state.update(state["model"]["trunk"])
        super().load_state_dict(current_state)

    def forward(self, x):
        """
        Perform computation of blocks in the order define in get_blocks.
        """
        raise NotImplementedError

    def extract_features(self, x):
        """
        Extract features from the model.

        Derived classes can implement this method to extract the features before
        applying the final fc layer.
        """
        return self.forward(x)

    def build_attachable_block(self, name, module):
        """
        Add a wrapper to the module to allow to attach heads to the module.
        """
        if name in self._attachable_blocks:
            raise ValueError("Found duplicated block name {}".format(name))
        block = ClassyBlock(name, module)
        self._attachable_blocks[name] = block
        return block

    @property
    def attachable_block_names(self):
        """
        Return names of all attachable blocks.
        """
        return self._attachable_blocks.keys()

    def _clear_heads(self):
        # clear all existing heads
        self._heads.clear()
        self._head_outputs.clear()

    def set_heads(self, heads: Dict[str, Dict[str, ClassyHead]]):
        """Attach all the heads to corresponding blocks.

        A head is expected to be a ClassyHead object. For more
        details, see :class:`ClassyHead`.

        Args:
            heads (Dict): a mapping between attachable block name
                and a dictionary of heads attached to that block. For
                example, if you have two different teams that want to
                attach two different heads for downstream classifiers to
                the 15th block, then they would use:
                heads = {"block15":
                    {"team1": classifier_head1, "team2": classifier_head2}
                }
        """
        self._clear_heads()

        head_ids = set()
        for block_name, block_heads in heads.items():
            if block_name not in self._attachable_blocks:
                raise ValueError(
                    "block {} does not exist or can not be attached".format(block_name)
                )
            self._attachable_blocks[block_name].set_cache_output()
            for head in block_heads.values():
                if head.unique_id in head_ids:
                    raise ValueError("head id {} already exists".format(head.unique_id))
                head_ids.add(head.unique_id)
            self._heads[block_name] = nn.ModuleDict(block_heads)

    def get_heads(self):
        """Returns the heads on the model

        Function returns the heads a dictionary of block names to
        nn.modules attached to that block.

        """
        return {block_name: dict(heads) for block_name, heads in self._heads.items()}

    @property
    def head_outputs(self):
        """Return outputs of all heads in the format of Dict[head_id, output]

        Head outputs are cached during a forward pass.
        """
        return self._head_outputs.copy()

    def get_block_outputs(self) -> Dict[str, torch.Tensor]:
        outputs = {}
        for name, block in self._attachable_blocks.items():
            outputs[name] = block.output
        return outputs

    def execute_heads(self) -> Dict[str, torch.Tensor]:
        block_outs = self.get_block_outputs()
        outputs = {}
        for block_name, heads in self._heads.items():
            for head in heads.values():
                outputs[head.unique_id] = head(block_outs[block_name])
        self._head_outputs = outputs
        return outputs

    def get_optimizer_params(self, bn_weight_decay=False):
        """Returns param groups for optimizer.

        Function to return dict of params with "keys" from
        {"regularized_params", "unregularized_params"}
        to "values" a list of torch Params.

        "weight_decay" provided as part of optimizer is only used
        for "regularized_params". For "unregularized_params", weight_decay is set
        to 0.0

        This implementation sets BatchNorm's all trainable params to be
        unregularized_params if bn_weight_decay is False.

        Override this function for any custom behavior.

        Args:
            bn_weight_decay (bool): Apply weight decay to bn params if true
        """
        unregularized_params = []
        regularized_params = []
        for module in self.modules():
            # If module has children (i.e. internal node of constructed DAG) then
            # only add direct parameters() to the list of params, else go over
            # children node to find if they are BatchNorm or have "bias".
            if list(module.children()) != []:
                for params in module.parameters(recurse=False):
                    if params.requires_grad:
                        regularized_params.append(params)
            elif not bn_weight_decay and isinstance(
                module,
                (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm),
            ):
                for params in module.parameters():
                    if params.requires_grad:
                        unregularized_params.append(params)
            else:
                for params in module.parameters():
                    if params.requires_grad:
                        regularized_params.append(params)
        return {
            "regularized_params": regularized_params,
            "unregularized_params": unregularized_params,
        }

    @property
    def input_shape(self):
        """If implemented, returns expected input tensor shape
        """
        raise NotImplementedError

    @property
    def output_shape(self):
        """If implemented, returns expected output tensor shape
        """
        raise NotImplementedError

    @property
    def model_depth(self):
        """If implemented, returns number of layers in model
        """
        raise NotImplementedError

    @property
    def evaluation_mode(self):
        """Used by video models for averaging over contiguous clips.

        TODO: Remove this once we have a video task, this logic should
        live in a video specific task

        """
        return ClassyModelEvaluationMode.DEFAULT



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Optional

from classy_vision import tasks
from classy_vision.generic.distributed_util import is_master
from classy_vision.hooks.classy_hook import ClassyHook


try:
    import progressbar

    progressbar_available = True
except ImportError:
    progressbar_available = False


class ProgressBarHook(ClassyHook):
    """
    Displays a progress bar to show progress in processing batches.
    """

    on_rendezvous = ClassyHook._noop
    on_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(self) -> None:
        """The constructor method of ProgressBarHook."""
        super().__init__()
        self.progress_bar: Optional[progressbar.ProgressBar] = None
        self.bar_size: int = 0
        self.batches: int = 0

    def on_phase_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Create and display a progress bar with 0 progress."""
        if not progressbar_available:
            raise RuntimeError(
                "progressbar module not installed, cannot use ProgressBarHook"
            )

        if is_master():
            self.bar_size = task.num_batches_per_phase
            self.batches = 0
            self.progress_bar = progressbar.ProgressBar(self.bar_size)
            self.progress_bar.start()

    def on_update(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Update the progress bar with the batch size."""
        if is_master() and self.progress_bar is not None:
            self.batches += 1
            self.progress_bar.update(min(self.batches, self.bar_size))

    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Clear the progress bar at the end of the phase."""
        if is_master() and self.progress_bar is not None:
            self.progress_bar.finish()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.optim
from classy_vision.generic.util import is_pos_float
from classy_vision.optim.param_scheduler import build_param_scheduler

from . import ClassyOptimizer, register_optimizer
from .param_scheduler.classy_vision_param_scheduler import ClassyParamScheduler


@register_optimizer("rmsprop")
class RMSProp(ClassyOptimizer):
    def __init__(
        self,
        lr_scheduler: ClassyParamScheduler,
        momentum: float,
        weight_decay: float,
        alpha: float,
        eps: float = 1e-8,
        centered: bool = False,
    ) -> None:
        super().__init__(lr_scheduler=lr_scheduler)

        self.momentum = momentum
        self.weight_decay = weight_decay
        self.alpha = alpha
        self.eps = eps
        self.centered = centered

    def init_pytorch_optimizer(self, model):
        super().init_pytorch_optimizer(model)
        self.optimizer = torch.optim.RMSprop(
            self.param_groups_override,
            lr=self.lr,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
            alpha=self.alpha,
            eps=self.eps,
            centered=self.centered,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "RMSProp":
        """Instantiates a RMSProp from a configuration.

        Args:
            config: A configuration for a RMSProp.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A RMSProp instance.
        """
        # Default params
        config.setdefault("eps", 1e-8)
        config.setdefault("centered", False)

        assert (
            "lr" in config
        ), "Config must contain a learning rate 'lr' section for RMSProp optimizer"
        for key in ["momentum", "alpha"]:
            assert (
                key in config
                and config[key] >= 0.0
                and config[key] < 1.0
                and type(config[key]) == float
            ), f"Config must contain a '{key}' in [0, 1) for RMSProp optimizer"
        for key in ["weight_decay", "eps"]:
            assert key in config and is_pos_float(
                config[key]
            ), f"Config must contain a positive '{key}' for RMSProp optimizer"
        assert "centered" in config and isinstance(
            config["centered"], bool
        ), "Config must contain a boolean 'centered' param for RMSProp optimizer"

        lr_config = config["lr"]
        if not isinstance(lr_config, dict):
            lr_config = {"name": "constant", "value": lr_config}

        lr_config["num_epochs"] = config["num_epochs"]
        lr_scheduler = build_param_scheduler(lr_config)

        return cls(
            lr_scheduler=lr_scheduler,
            momentum=config["momentum"],
            weight_decay=config["weight_decay"],
            alpha=config["alpha"],
            eps=config["eps"],
            centered=config["centered"],
        )

    @property
    def parameters(self) -> Dict[str, Any]:
        return {
            "lr": self.lr,
            "momentum": self.momentum,
            "weight_decay": self.weight_decay,
            "alpha": self.alpha,
            "eps": self.eps,
            "centered": self.centered,
        }



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from . import ClassyParamScheduler, register_param_scheduler


@register_param_scheduler("constant")
class ConstantParamScheduler(ClassyParamScheduler):
    """
    Returns a constant value for a optimizer param.
    """

    def __init__(self, value: float):
        super().__init__()
        self._value = value

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ConstantParamScheduler":
        """Instantiates a ConstantParamScheduler from a configuration.

        Args:
            config: A configuration for a ConstantParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A ConstantParamScheduler instance.
        """
        assert "value" in config
        return cls(value=config["value"])

    def __call__(self, where: float):
        if where >= 1.0:
            raise RuntimeError(f"Invalid where parameter for scheduler: {where}")

        return self._value



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict, Optional

from classy_vision import tasks
from classy_vision.generic.distributed_util import get_rank
from classy_vision.hooks.classy_hook import ClassyHook


class LossLrMeterLoggingHook(ClassyHook):
    """
    Logs the loss, optimizer LR, and meters. Logs at the end of a phase.
    """

    on_rendezvous = ClassyHook._noop
    on_start = ClassyHook._noop
    on_phase_start = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(self, log_freq: Optional[int] = None) -> None:
        """The constructor method of LossLrMeterLoggingHook.

        Args:
            log_freq: if specified, logs every log_freq batches also.

        """
        super().__init__()
        self.log_freq: Optional[int] = log_freq

    def on_loss_and_meter(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Log metrics every log_freq batches, if log_freq is not None.
        """
        if self.log_freq is None:
            return
        batches = len(task.losses)
        if batches and batches % self.log_freq == 0:
            logging.info("Local unsynced metric values:")
            self._log_loss_meters(task, local_variables)

    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Log the loss, optimizer LR, and meters for the phase.
        """
        batches = len(task.losses)
        if batches:
            # Most trainers will sync meters on phase end, however we
            # do not explicitly state this since it is possible for a
            # trainer to implement an unsynced end of phase meter or
            # for meters to not provide a sync function.
            logging.info("End of phase metric values:")
            self._log_loss_meters(task, local_variables)
            if task.train:
                self._log_lr(task, local_variables)

    def on_update(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Log the LR every log_freq batches, if log_freq is not None.
        """
        if self.log_freq is None:
            return
        batches = len(task.losses)
        if batches and batches % self.log_freq == 0:
            self._log_lr(task, local_variables)

    def _log_lr(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Compute and log the optimizer LR.
        """
        optimizer_lr = task.optimizer.lr
        logging.info("Learning Rate: {}\n".format(optimizer_lr))

    def _log_loss_meters(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """
        Compute and log the loss and meters.
        """

        phase_type = task.phase_type
        phase_type_idx = task.train_phase_idx if task.train else task.eval_phase_idx
        batches = len(task.losses)

        # Loss for the phase
        loss = sum(task.losses) / (batches * task.get_batchsize_per_replica())

        log_strs = [
            "Rank: {}, {} phase: {}, processed batches: {}".format(
                get_rank(), phase_type, phase_type_idx, batches
            ),
            "{} loss: {}".format(phase_type, loss),
            "Meters:",
        ]
        for meter in task.meters:
            log_strs.append("{}".format(meter))
        logging.info("\n".join(log_strs))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.step_scheduler import StepParamScheduler


class TestStepScheduler(unittest.TestCase):
    _num_epochs = 12

    def _get_valid_config(self):
        return {
            "name": "step",
            "num_epochs": self._num_epochs,
            "values": [0.1, 0.01, 0.001, 0.0001],
        }

    def test_invalid_config(self):
        # Invalid num epochs
        config = self._get_valid_config()

        bad_config = copy.deepcopy(config)
        bad_config["num_epochs"] = -1
        with self.assertRaises(AssertionError):
            StepParamScheduler.from_config(bad_config)

        # Invalid Values
        bad_config["num_epochs"] = config["num_epochs"]
        del bad_config["values"]
        with self.assertRaises(AssertionError):
            StepParamScheduler.from_config(bad_config)

        bad_config["values"] = {"a": "b"}
        with self.assertRaises(AssertionError):
            StepParamScheduler.from_config(bad_config)

        bad_config["values"] = []
        with self.assertRaises(AssertionError):
            StepParamScheduler.from_config(bad_config)

    def test_scheduler(self):
        config = self._get_valid_config()

        scheduler = StepParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [
            0.1,
            0.1,
            0.1,
            0.01,
            0.01,
            0.01,
            0.001,
            0.001,
            0.001,
            0.0001,
            0.0001,
            0.0001,
        ]

        self.assertEqual(schedule, expected_schedule)

    def test_build_step_scheduler(self):
        config = self._get_valid_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, StepParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
import unittest
import unittest.mock as mock

import torch
import torch.nn as nn
from classy_vision.hooks import ExponentialMovingAverageModelHook
from classy_vision.models import ClassyModel


class TestModel(ClassyModel):
    def __init__(self):
        super().__init__()
        self.fc = nn.Linear(10, 10)
        self.bn = nn.BatchNorm1d(10)

    def init_fc_weight(self):
        nn.init.zeros_(self.fc.weight)

    def update_fc_weight(self):
        nn.init.ones_(self.fc.weight)

    def forward(self, x):
        return self.bn(self.fc(x))


class TestExponentialMovingAverageModelHook(unittest.TestCase):
    def _map_device_string(self, device):
        return "cuda" if device == "gpu" else "cpu"

    def _test_exponential_moving_average_hook(self, model_device, hook_device):
        task = mock.MagicMock()
        model = TestModel().to(device=self._map_device_string(model_device))
        local_variables = {}
        task.base_model = model
        task.train = True
        decay = 0.5
        num_updates = 10
        model.init_fc_weight()
        exponential_moving_average_hook = ExponentialMovingAverageModelHook(
            decay=decay, device=hook_device
        )

        exponential_moving_average_hook.on_start(task, local_variables)
        exponential_moving_average_hook.on_phase_start(task, local_variables)
        # set the weights to all ones and simulate 10 updates
        task.base_model.update_fc_weight()
        fc_weight = model.fc.weight.clone()
        for _ in range(num_updates):
            exponential_moving_average_hook.on_update(task, local_variables)
        exponential_moving_average_hook.on_phase_end(task, local_variables)
        # the model weights shouldn't have changed
        self.assertTrue(torch.allclose(model.fc.weight, fc_weight))

        # simulate a test phase now
        task.train = False
        exponential_moving_average_hook.on_phase_start(task, local_variables)
        exponential_moving_average_hook.on_phase_end(task, local_variables)

        # the model weights should be updated to the ema weights
        self.assertTrue(
            torch.allclose(
                model.fc.weight, fc_weight * (1 - math.pow(1 - decay, num_updates))
            )
        )

        # simulate a train phase again
        task.train = True
        exponential_moving_average_hook.on_phase_start(task, local_variables)

        # the model weights should be back to the old value
        self.assertTrue(torch.allclose(model.fc.weight, fc_weight))

    def test_get_model_state_iterator(self):
        device = "gpu" if torch.cuda.is_available() else "cpu"
        model = TestModel().to(device=self._map_device_string(device))
        decay = 0.5
        # test that we pick up the right parameters in the iterator
        for consider_bn_buffers in [True, False]:
            exponential_moving_average_hook = ExponentialMovingAverageModelHook(
                decay=decay, consider_bn_buffers=consider_bn_buffers, device=device
            )
            iterable = exponential_moving_average_hook.get_model_state_iterator(model)
            fc_found = False
            bn_found = False
            bn_buffer_found = False
            for _, param in iterable:
                if any(param is item for item in model.fc.parameters()):
                    fc_found = True
                if any(param is item for item in model.bn.parameters()):
                    bn_found = True
                if any(param is item for item in model.bn.buffers()):
                    bn_buffer_found = True
            self.assertTrue(fc_found)
            self.assertTrue(bn_found)
            self.assertEqual(bn_buffer_found, consider_bn_buffers)

    def test_exponential_moving_average_hook(self):
        device = "gpu" if torch.cuda.is_available() else "cpu"
        self._test_exponential_moving_average_hook(device, device)

    @unittest.skipUnless(torch.cuda.is_available(), "This test needs a gpu to run")
    def test_mixed_devices(self):
        """Tests that the hook works when the model and hook's device are different"""
        self._test_exponential_moving_average_hook("cpu", "gpu")
        self._test_exponential_moving_average_hook("gpu", "cpu")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import random
import unittest

import numpy
import torch
import torchvision.transforms as transforms
from classy_vision.dataset.core.random_image_datasets import (
    RandomImageBinaryClassDataset,
    SampleType,
)
from classy_vision.dataset.transforms import build_transforms
from classy_vision.dataset.transforms.util import (
    GenericImageTransform,
    ImagenetAugmentTransform,
    ImagenetNoAugmentTransform,
    build_field_transform_default_imagenet,
)


class DatasetTransformsUtilTest(unittest.TestCase):
    def get_test_image_dataset(self, sample_type):
        return RandomImageBinaryClassDataset(
            crop_size=224,
            class_ratio=0.5,
            num_samples=100,
            seed=0,
            sample_type=sample_type,
        )

    def transform_checks(
        self, sample, transform, expected_transform, key, transformed_key=None
    ):
        # If transformed key is None, then use key
        transformed_key = transformed_key if transformed_key is not None else key
        input_image = copy.deepcopy(sample[key])

        torch.manual_seed(0)
        numpy.random.seed(0)
        random.seed(0)
        output_image = transform(sample)[transformed_key]

        torch.manual_seed(0)
        numpy.random.seed(0)
        random.seed(0)
        self.assertTrue(torch.allclose(output_image, expected_transform(input_image)))

    def test_build_dict_field_transform_default_imagenet(self):
        dataset = self.get_test_image_dataset(SampleType.DICT)

        # should apply the transform in the config
        config = [{"name": "ToTensor"}]
        default_transform = transforms.Compose(
            [transforms.CenterCrop(100), transforms.ToTensor()]
        )
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform
        )
        sample = dataset[0]
        self.transform_checks(sample, transform, transforms.ToTensor(), "input")

        # should apply default_transform
        config = None
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform
        )
        sample = dataset[0]
        self.transform_checks(sample, transform, default_transform, "input")

        # should apply the transform for a test split
        transform = build_field_transform_default_imagenet(config, split="test")
        sample = dataset[0]
        self.transform_checks(sample, transform, ImagenetNoAugmentTransform(), "input")

    def test_build_tuple_field_transform_default_imagenet(self):
        dataset = self.get_test_image_dataset(SampleType.TUPLE)

        # should apply the transform in the config
        config = [{"name": "ToTensor"}]
        default_transform = transforms.Compose(
            [transforms.CenterCrop(100), transforms.ToTensor()]
        )
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform, key=0, key_map_transform=None
        )
        sample = dataset[0]
        self.transform_checks(sample, transform, transforms.ToTensor(), 0)

        # should apply default_transform
        config = None
        transform = build_field_transform_default_imagenet(
            config, default_transform=default_transform, key=0, key_map_transform=None
        )
        sample = dataset[0]
        self.transform_checks(sample, transform, default_transform, 0)

        # should apply the transform for a test split
        transform = build_field_transform_default_imagenet(
            config, split="test", key=0, key_map_transform=None
        )
        sample = dataset[0]
        self.transform_checks(sample, transform, ImagenetNoAugmentTransform(), 0)

    def test_apply_transform_to_key_from_config(self):
        dataset = self.get_test_image_dataset(SampleType.DICT)

        config = [
            {
                "name": "apply_transform_to_key",
                "transforms": [{"name": "ToTensor"}],
                "key": "input",
            }
        ]
        transform = build_transforms(config)
        sample = dataset[0]
        self.transform_checks(sample, transform, transforms.ToTensor(), "input")

    def test_generic_image_transform(self):
        dataset = self.get_test_image_dataset(SampleType.TUPLE)

        # Check class constructor
        transform = GenericImageTransform(transform=transforms.ToTensor())
        sample = dataset[0]
        self.transform_checks(sample, transform, transforms.ToTensor(), 0, "input")

        transform = GenericImageTransform(split="train")
        sample = dataset[0]
        self.transform_checks(sample, transform, ImagenetAugmentTransform(), 0, "input")

        transform = GenericImageTransform(split="test")
        sample = dataset[0]
        self.transform_checks(
            sample, transform, ImagenetNoAugmentTransform(), 0, "input"
        )

        # Check from_config constructor / registry
        config = [
            {"name": "generic_image_transform", "transforms": [{"name": "ToTensor"}]}
        ]
        transform = build_transforms(config)
        sample = dataset[0]
        self.transform_checks(sample, transform, transforms.ToTensor(), 0, "input")

        # Check with Imagenet defaults
        config = [{"name": "generic_image_transform", "split": "train"}]
        transform = build_transforms(config)
        sample = dataset[0]
        self.transform_checks(sample, transform, ImagenetAugmentTransform(), 0, "input")

        config = [{"name": "generic_image_transform", "split": "test"}]
        transform = build_transforms(config)
        sample = dataset[0]
        self.transform_checks(
            sample, transform, ImagenetNoAugmentTransform(), 0, "input"
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import multiprocessing
import os
from typing import Any, Callable, Dict, Optional

import torch
from classy_vision.generic.distributed_util import get_rank, get_world_size
from torch.utils.data import Sampler
from torchvision.datasets.samplers.clip_sampler import (
    DistributedSampler,
    RandomClipSampler,
    UniformClipSampler,
)

from .classy_dataset import ClassyDataset


class MaxLengthClipSampler(Sampler):
    """MaxLengthClipSampler is a thin wrapper on top of clip samplers in TorchVision.

    It takes as input a TorchVision clip sampler, and an optional argument
    `num_samples` to limit the number of samples.
    """

    def __init__(self, clip_sampler, num_samples=None):
        """The constructor method of MaxLengthClipSampler.

        Args:
            clip_sampler: clip sampler without a limit on the total number of clips
                it can sample, such as RandomClipSampler and UniformClipSampler.
            num_samples: if provided, it denotes the maximal number of clips the sampler
                will return

        """
        self.clip_sampler = clip_sampler
        self.num_samples = num_samples

    def __iter__(self):
        num_samples = len(self)
        n = 0
        for clip in self.clip_sampler:
            if n < num_samples:
                yield clip
                n += 1
            else:
                break

    def __len__(self):
        full_size = len(self.clip_sampler)
        if self.num_samples is None:
            return full_size

        return min(full_size, self.num_samples)


class ClassyVideoDataset(ClassyDataset):
    """Interface specifying what a ClassyVision video dataset is expected to provide.
    """

    def __init__(
        self,
        dataset: Any,
        split: str,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Callable,
        num_samples: Optional[int],
        clips_per_video: int,
    ):
        """The constructor method of ClassyVideoDataset.

        Args:
            dataset: the underlying video dataset from either TorchVision or other
                source. It should have an attribute `video_clips` of type
                torchvision.datasets.video_utils.VideoClips
            split: dataset split. Must be either "train" or "test"
            batchsize_per_replica: batch size per model replica
            shuffle: If true, shuffle video clips.
            transform: callable function to transform video clip sample from
                ClassyVideoDataset
            num_samples: If provided, return at most `num_samples` video clips
            clips_per_video: The number of clips sampled from each video

        """
        super(ClassyVideoDataset, self).__init__(
            dataset, split, batchsize_per_replica, shuffle, transform, num_samples
        )
        # Assignments:
        self.clips_per_video = clips_per_video

    @classmethod
    def parse_config(cls, config: Dict[str, Any]):
        """Parse config to prepare arguments needed by the class constructor."""
        assert "frames_per_clip" in config, "frames_per_clip must be set"

        video_width = config.get("video_width", 0)
        video_height = config.get("video_height", 0)
        video_min_dimension = config.get("video_min_dimension", 0)
        audio_samples = config.get("audio_samples", 0)
        step_between_clips = config.get("step_between_clips", 1)
        frame_rate = config.get("frame_rate", None)
        clips_per_video = config.get("clips_per_video", 1)

        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = super().parse_config(config)

        if not config["split"] == "train":
            # At testing time, we do not crop frames but conduct a FCN-style evaluation.
            # Video spatial resolution can vary from video to video. So we test one
            # video at a time, and NO. of clips in a minibatch should be equal to
            # No. of clips sampled from a video
            if not batchsize_per_replica == clips_per_video:
                logging.warning(
                    f"For testing, batchsize per replica ({batchsize_per_replica})"
                    + f"should be equal to clips_per_video ({clips_per_video})"
                )

        return (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
            config["frames_per_clip"],
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            step_between_clips,
            frame_rate,
            clips_per_video,
        )

    @classmethod
    def load_metadata(
        cls,
        filepath: str,
        video_dir: Optional[str] = None,
        update_file_path: bool = False,
    ) -> Dict[str, Any]:
        """Load pre-computed video dataset meta data.

        Video dataset meta data computation takes minutes on small dataset and hours
        on large dataset, and thus is time-consuming. However, it only needs to be
        computed once, and can be saved into file. Later we can load the meta data
        to reuse it.

        The format of meta data is defined in TorchVision as shown below.
        https://github.com/pytorch/vision/blob/master/torchvision/datasets/
        video_utils.py#L131

        For each video, meta data contains the video file path, presentation
        timestamps of all video frames, and video fps.

        Args:
            filepath: file path of pre-computed meta data
            video_dir: If provided, the folder where video files are stored.
            update_file_path: If true, replace the directory part of video file path
                in meta data with the actual video directory provided in `video_dir`.
                This is needed for successsfully reusing pre-computed meta data
                when video directory has been moved and it is no longer consitent
                with the full video file path saved in the meta data.
        """
        metadata = torch.load(filepath)
        if video_dir is not None and update_file_path:
            # video path in meta data can be computed in a different root video folder
            # If we use a different root video folder, we need to update the video paths
            assert os.path.exists(video_dir), "folder does not exist: %s" % video_dir
            for idx, video_path in enumerate(metadata["video_paths"]):
                # video path template is $VIDEO_DIR/$CLASS_NAME/$VIDEO_FILE
                dirname, filename = os.path.split(video_path)
                _, class_name = os.path.split(dirname)
                metadata["video_paths"][idx] = os.path.join(
                    video_dir, class_name, filename
                )
        return metadata

    @classmethod
    def save_metadata(cls, metadata: Dict[str, Any], filepath: str):
        """Save dataset meta data into a file.

        Args:
            metadata: dataset meta data, which contains video meta infomration, such
                as video file path, video fps, video frame timestamp in each video.
                For the format of dataset meta data, check the TorchVision
                documentations below.
                https://github.com/pytorch/vision/blob/master/torchvision/datasets
                /video_utils.py#L132-L137

            filepath: file path where the meta data will be saved

        """
        filedir = os.path.dirname(filepath)
        if not os.path.exists(filedir):
            try:
                os.mkdirs(filedir)
            except Exception as err:
                logging.warn(f"Fail to create folder: {filedir}")
                raise err
        logging.info(f"Save metadata to file: {filedir}")
        try:
            torch.save(metadata, filepath)
        except ValueError:
            logging.warn(f"Fail to save metadata to file: {filepath}")

    @property
    def video_clips(self):
        """Attribute video_clips.

        It is used in `_get_sampler` method. Its data type should be
            torchvision.datasets.video_utils.VideoClips.
        """
        return self.dataset.video_clips

    def _get_sampler(self, epoch) -> "DistributedSampler":
        if self.split == "train":
            # For video model training, we don't necessarily want to use all possible
            # clips in the video in one training epoch. More often, we randomly
            # sample at most N clips per training video. In practice, N is often 1
            clip_sampler = RandomClipSampler(self.video_clips, self.clips_per_video)
        else:
            # For video model testing, we sample N evenly spaced clips per test
            # video. We will simply average predictions over them
            clip_sampler = UniformClipSampler(self.video_clips, self.clips_per_video)
        clip_sampler = MaxLengthClipSampler(clip_sampler, num_samples=self.num_samples)
        world_size = get_world_size()
        rank = get_rank()
        sampler = DistributedSampler(
            clip_sampler,
            num_replicas=world_size,
            rank=rank,
            shuffle=self.shuffle,
            group_size=self.clips_per_video,
        )
        sampler.set_epoch(epoch)
        return sampler

    def iterator(self, *args, **kwargs):
        """Overrides the implementation in parent class `ClassyDataset`.

        You can check all the usable positional and keyword arguments in parent
        class `ClassyDataset.iterator(...)`.
        For video dataset, it may use VideoClips class from TorchVision,
        which may use a cpp python extension for video decoding when video backend
        is set to `video_reader`. In such case, it is difficult to use "spawning"
        as multiprocessing start method. Thus we choose "fork" as multiprocessing
        start method.
        """
        if "num_workers" in kwargs and kwargs["num_workers"] > 0:
            mp = multiprocessing.get_context("fork")
            kwargs["multiprocessing_context"] = mp
        return super(ClassyVideoDataset, self).iterator(*args, **kwargs)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any, Callable, Dict, Optional, Union

from classy_vision.dataset import ClassyDataset, register_dataset
from classy_vision.dataset.core.random_image_datasets import (
    RandomImageBinaryClassDataset,
    SampleType,
)
from classy_vision.dataset.transforms import ClassyTransform, build_transforms


@register_dataset("my_dataset")
class MyDataset(ClassyDataset):
    def __init__(
        self,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: int,
        crop_size: int,
        class_ratio: float,
        seed: int,
        split: Optional[str] = None,
    ) -> None:
        dataset = RandomImageBinaryClassDataset(
            crop_size, class_ratio, num_samples, seed, SampleType.TUPLE
        )
        super().__init__(
            dataset, split, batchsize_per_replica, shuffle, transform, num_samples
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "MyDataset":
        assert all(key in config for key in ["crop_size", "class_ratio", "seed"])

        split = config.get("split")
        crop_size = config["crop_size"]
        class_ratio = config["class_ratio"]
        seed = config["seed"]
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = cls.parse_config(config)
        transform = build_transforms(transform_config)
        return cls(
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            crop_size,
            class_ratio,
            seed,
            split=split,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math

import numpy as np
import torch
import torch.nn.modules as nn
from classy_vision.generic.util import get_model_dummy_input, is_pos_int
from PIL import Image


try:
    import visdom
    from tensorboardX import SummaryWriter
except ImportError:
    pass

# define loss types:
vis = []  # using a list makes this work as an upvalue
UNSUPPORTED_LOSSES = (
    nn.CosineEmbeddingLoss,
    nn.PoissonNLLLoss,
    nn.KLDivLoss,
    nn.TripletMarginLoss,
)
REGRESSION_LOSSES = (nn.L1Loss, nn.SmoothL1Loss, nn.MSELoss)


# connection to visdom:
def visdom_connect(server=None, port=None):
    """Connects to a visdom server if not currently connected."""
    if not visdom_connected():
        vis.append(visdom.Visdom(server=server, port=port))


# check if we are connected to visdom:
def visdom_connected():
    """Returns True if the client is connected to a visdom server."""
    return (
        len(vis) > 0
        and hasattr(vis[-1], "check_connection")
        and vis[-1].check_connection()
    )


# function that plots learning curve:
def plot_learning_curves(curves, visdom_server=None, env=None, win=None, title=""):
    """Plots the specified dict of learning curves in visdom. Optionally, the
    environment, window handle, and title for the visdom plot can be specified.
    """

    if visdom_server is None and visdom_connected():
        visdom_server = vis[-1]

    # return if we are not connected to visdom server:
    if not visdom_server or not visdom_server.check_connection():
        print("WARNING: Not connected to visdom. Skipping plotting.")
        return

    # assertions:
    assert type(curves) == dict
    assert all(type(curve) == list for _, curve in curves.items())

    # remove batch time curves:
    _curves = {k: curves[k] for k in curves.keys() if "batch time" not in k}

    # show plot:
    X = torch.stack([torch.FloatTensor(curve) for _, curve in _curves.items()], dim=1)
    Y = torch.arange(0, X.size(0))
    Y = Y.view(Y.numel(), 1).expand(Y.numel(), X.size(1))
    opts = {"title": title, "legend": list(_curves.keys()), "xlabel": "Epochs"}
    return visdom_server.line(X, Y, env=env, win=win, opts=opts)


# function that plots loss functions:
def plot_losses(losses, visdom_server=None, env=None, win=None, title=""):
    """Constructs a plot of specified losses as function of y * f(x). The losses
    are a list of nn.Module losses. Optionally, the environment, window handle,
    and title for the visdom plot can be specified.
    """

    if visdom_server is None and visdom_connected():
        visdom_server = vis[-1]

    # return if we are not connected to visdom server:
    if not visdom_server or not visdom_server.check_connection():
        print("WARNING: Not connected to visdom. Skipping plotting.")
        return

    # assertions:
    if isinstance(losses, nn.Module):
        losses = [losses]
    assert type(losses) == list
    assert all(isinstance(loss, nn.Module) for loss in losses)
    if any(isinstance(loss, UNSUPPORTED_LOSSES) for loss in losses):
        raise NotImplementedError("loss function not supported")

    # loop over all loss functions:
    for idx, loss in enumerate(losses):

        # construct scores and targets:
        score = torch.arange(-5.0, 5.0, 0.005)
        if idx == 0:
            loss_val = torch.FloatTensor(score.size(0), len(losses))
        if isinstance(loss, REGRESSION_LOSSES):
            target = torch.FloatTensor(score.size()).fill_(0.0)
        else:
            target = torch.LongTensor(score.size()).fill_(1)

        # compute loss values:
        for n in range(0, score.nelement()):
            loss_val[n][idx] = loss(
                score.narrow(0, n, 1), target.narrow(0, n, 1)
            ).item()

    # show plot:
    title = str(loss) if title == "" else title
    legend = [str(loss) for loss in losses]
    opts = {"title": title, "xlabel": "Score", "ylabel": "Loss", "legend": legend}
    win = visdom_server.line(loss_val, score, env=env, win=win, opts=opts)
    return win


def plot_model(
    model, size=(3, 224, 224), input_key=None, writer=None, folder="", train=True
):
    """Visualizes a model in TensorBoard.

    The TensorBoard writer can be either specified directly via `writer` or can
    be specified via a `folder`.

    The model can be run in training or evaluation model via the `train` argument.

    Example usage on devserver:
     - Install TensorBoard using: `sudo feature install tensorboard`
     - Start TensorBoard using: `tensorboard --port=8098 --logdir <folder>`
    """

    assert (
        writer is not None or folder != ""
    ), "must specify SummaryWriter or folder to create SummaryWriter in"
    input = get_model_dummy_input(model, size, input_key)
    if writer is None:
        writer = SummaryWriter(log_dir=folder, comment="Model graph")
    with writer:
        orig_train = model.training
        model.train(train)  # visualize model in desired mode
        writer.add_graph(model, input_to_model=(input,))
        model.train(orig_train)


# function that produces an image map:
def image_map(
    mapcoord, dataset, mapsize=5000, imsize=32, unnormalize=None, snap_to_grid=False
):
    """Constructs a 2D map of images.

    The 2D coordinates for each of the images are specified in `mapcoord`, the
    corresponding images are in `dataset`. Optional arguments set the size of
    the map images, the size of the images themselves, the unnormalization
    transform, and whether or not to snap images to a grid.
    """

    # assertions:
    if type(mapcoord) == np.ndarray:
        mapcoord = torch.from_numpy(mapcoord)
    assert torch.is_tensor(mapcoord)
    if isinstance(dataset, torch.utils.data.dataloader.DataLoader):
        dataset = dataset.dataset
    assert isinstance(dataset, torch.utils.data.dataset.Dataset)
    assert is_pos_int(mapsize)
    assert is_pos_int(imsize)
    if unnormalize is not None:
        assert callable(unnormalize)

    # initialize some variables:
    import torchvision.transforms.functional as F

    background = 255
    mapim = torch.ByteTensor(3, mapsize, mapsize).fill_(background)

    # normalize map coordinates:
    mapc = mapcoord.add(-mapcoord.min())
    mapc.div_(mapc.max())

    # loop over images:
    for idx in range(len(dataset)):

        # compute grid location:
        if snap_to_grid:
            y = 1 + int(math.floor(mapc[idx][0] * (mapsize - imsize - 2)))
            x = 1 + int(math.floor(mapc[idx][1] * (mapsize - imsize - 2)))
        else:
            y = 1 + int(math.floor(mapc[idx][0] * (math.floor(mapsize - imsize) - 2)))
            x = 1 + int(math.floor(mapc[idx][1] * (math.floor(mapsize - imsize) - 2)))

        # check whether we can overwrite this location:
        overwrite = not snap_to_grid
        if not overwrite:
            segment = mapim.narrow(1, y, imsize).narrow(2, x, imsize)
            overwrite = segment.eq(background).all()

        # draw image:
        if overwrite:

            # load, unnormalize, and resize image:
            image = dataset[idx][0]
            if unnormalize is not None:
                image = unnormalize(image)
            resized_im = F.to_tensor(
                F.resize(F.to_pil_image(image), imsize, Image.BILINEAR)
            )

            # place image:
            segment = mapim.narrow(1, y, imsize).narrow(2, x, imsize)
            segment.copy_(resized_im.mul_(255.0).byte())

    # return map:
    return mapim



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import shutil
import tempfile
import unittest
from test.generic.config_utils import get_test_task_config

import torch
import torch.nn as nn

from classy_vision.generic.util import load_checkpoint
from classy_vision.hooks import CheckpointHook
from classy_vision.models import ClassyModel, register_model
from classy_vision.tasks import build_task


@register_model("my_test_model")
class MyTestModel(ClassyModel):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 5)

    def forward(self, x):
        return self.linear(x)

    @classmethod
    def from_config(cls, config):
        return cls()


class TestClassyModel(unittest.TestCase):
    def setUp(self) -> None:
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.base_dir)

    def test_from_checkpoint(self):
        config = get_test_task_config()
        config["model"] = {"name": "my_test_model"}
        task = build_task(config)
        task.prepare()

        local_variables = {}
        checkpoint_folder = self.base_dir + "/checkpoint_end_test/"
        device = "cpu"
        input_args = {"config": config}

        # Simulate training by setting the model parameters to zero
        for param in task.model.parameters():
            param.data.zero_()

        checkpoint_hook = CheckpointHook(
            checkpoint_folder, input_args, phase_types=["train"]
        )

        # Create checkpoint dir, save checkpoint
        os.mkdir(checkpoint_folder)
        checkpoint_hook.on_start(task, local_variables)

        task.train = True
        checkpoint_hook.on_phase_end(task, local_variables)

        # Model should be checkpointed. load and compare
        checkpoint = load_checkpoint(checkpoint_folder, device)

        model = ClassyModel.from_checkpoint(checkpoint)
        self.assertTrue(isinstance(model, MyTestModel))

        # All parameters must be zero
        for param in model.parameters():
            self.assertTrue(torch.all(param.data == 0))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules

from .classy_head import ClassyHead


FILE_ROOT = Path(__file__).parent


HEAD_REGISTRY = {}
HEAD_CLASS_NAMES = set()


def register_head(name):
    """Registers a ClassyHead subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyHead from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyHead subclass, like this:

    |    @register_head("my_head")
    |    class MyHead(ClassyHead):
    |        ...

    To instantiate a head from a configuration file, see
    :func:`build_head`."""

    def register_head_cls(cls):
        if name in HEAD_REGISTRY:
            raise ValueError("Cannot register duplicate head ({})".format(name))
        if not issubclass(cls, ClassyHead):
            raise ValueError(
                "Head ({}: {}) must extend ClassyHead".format(name, cls.__name__)
            )
        if cls.__name__ in HEAD_CLASS_NAMES:
            raise ValueError(
                "Cannot register head with duplicate class name ({})".format(
                    cls.__name__
                )
            )
        HEAD_REGISTRY[name] = cls
        HEAD_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_head_cls


def build_head(config):
    """Builds a ClassyHead from a config.

    This assumes a 'name' key in the config which is used to determine what
    head class to instantiate. For instance, a config `{"name": "my_head",
    "foo": "bar"}` will find a class that was registered as "my_head"
    (see :func:`register_head`) and call .from_config on it."""

    assert "name" in config, "Expect name in config"
    assert "unique_id" in config, "Expect a global unique id in config"
    assert config["name"] in HEAD_REGISTRY, "unknown head"
    name = config["name"]
    head_config = copy.deepcopy(config)
    del head_config["name"]
    return HEAD_REGISTRY[name].from_config(head_config)


# automatically import any Python files in the heads/ directory
import_all_modules(FILE_ROOT, "classy_vision.heads")

from .fully_connected_head import FullyConnectedHead  # isort:skip
from .fully_convolutional_linear_head import FullyConvolutionalLinearHead  # isort:skip
from .identity_head import IdentityHead  # isort:skip


__all__ = [
    "ClassyHead",
    "FullyConnectedHead",
    "FullyConvolutionalLinearHead",
    "IdentityHead",
    "build_head",
    "register_head",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Various test loaders."""

import os
import random
import unittest
from itertools import chain


def _circleci_parallelism(suite):
    """Allow for parallelism in CircleCI for speedier tests.."""
    if int(os.environ.get("CIRCLE_NODE_TOTAL", 0)) <= 1:
        # either not running on circleci, or we're not using parallelism.
        return suite
    # tests are automatically sorted by discover, so we will get the same ordering
    # on all hosts.
    total = int(os.environ["CIRCLE_NODE_TOTAL"])
    index = int(os.environ["CIRCLE_NODE_INDEX"])

    # right now each test is corresponds to a /file/. Certain files are slower than
    # others, so we want to flatten it
    tests = [testfile._tests for testfile in suite._tests]
    tests = list(chain.from_iterable(tests))
    random.Random(42).shuffle(tests)
    tests = [t for i, t in enumerate(tests) if i % total == index]
    return unittest.TestSuite(tests)


def unittests():
    """
    Short tests.

    Runs on CircleCI on every commit. Returns everything in the tests root directory.
    """
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover("test", pattern="*_test.py")
    test_suite = _circleci_parallelism(test_suite)
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(unittests())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum
from typing import Any, Dict


class UpdateInterval(Enum):
    """
    Enum for specifying update frequency for scheduler.

    Attributes:
        EPOCH (str): Update param before each epoch
        STEP (str): Update param before each optimizer step
    """

    EPOCH = "epoch"
    STEP = "step"


class ClassyParamScheduler(object):
    """
    Base class for Classy parameter schedulers.

    Attributes:
        update_interval: Specifies how often to update each parameter
            (before each epoch or each batch)
    """

    # To be used for comparisons with where
    WHERE_EPSILON = 1e-6

    def __init__(self, update_interval: UpdateInterval = UpdateInterval.EPOCH):
        """
        Constructor for ClassyParamScheduler

        Args:
            update_interval: Specifies the frequency of the param updates
        """
        self.update_interval = update_interval

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyParamScheduler":
        """Instantiates a ClassyParamScheduler from a configuration.

        Args:
            config: A configuration for the ClassyParamScheduler.

        Returns:
            A ClassyParamScheduler instance.
        """
        raise NotImplementedError

    def __call__(self, where: float):
        """
        Get the param for a given point at training.

        For Classy Vision we update params (such as LR rate) based on
        the percent progress of training completed. This allows a
        scheduler to be agnostic to the exact specifications of a
        particular run (e.g. 120 epochs vs 90 epochs).

        Args:
            where: A float in [0;1) that represents how far training has progressed

        """
        raise NotImplementedError("Param schedulers must override __call__")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import itertools
import logging
from typing import Any, Dict, Iterable, Tuple

import torch
import torch.nn as nn
from classy_vision.hooks import ClassyHook
from classy_vision.tasks import ClassyTask


class ExponentialMovingAverageModelHook(ClassyHook):
    """
    Hook which keeps a track of the exponential moving average (EMA) of the model's
    parameters and applies the EMA params to the model during the test phases.

    Saving the state in cpu will save gpu memory, but will make training slower since
    the model parameters will need to be moved to cpu before the averaging.

    Note: This hooks stores two additional copies of the model's parameters, which will
        increase memory usage significantly.
    """

    on_rendezvous = ClassyHook._noop
    on_sample = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_backward = ClassyHook._noop
    on_end = ClassyHook._noop

    def __init__(
        self, decay: float, consider_bn_buffers: bool = True, device: str = "cpu"
    ) -> None:
        """The constructor method of ExponentialMovingAverageModelHook.

        Args:
            decay: EMA decay factor, should be in [0, 1]. A decay of 0 corresponds to
                always using the latest value (no EMA) and a decay of 1 corresponds to
                not updating weights after initialization.
            consider_bn_buffers: Whether to apply EMA to batch norm buffers
            device: Device to store the model state.
        """
        super().__init__()
        assert 0 <= decay <= 1, "Decay should be between 0 and 1"
        assert device in ["cpu", "gpu"], "Device should be one of cpu or gpu"
        self.decay: int = decay
        self.consider_bn_buffers = consider_bn_buffers
        self.device = "cuda" if device == "gpu" else "cpu"
        self.state.model_state = {}
        self.state.ema_model_state = {}
        logging.info(
            f"{self.__class__.__name__} initialized with a decay of "
            f"{decay} on device {device}"
        )

    def get_model_state_iterator(self, model: nn.Module) -> Iterable[Tuple[str, Any]]:
        """Get an iterator over the model state to apply EMA to."""
        iterable = model.named_parameters()
        if self.consider_bn_buffers:
            # also add batch norm buffers to the list of state params to iterate over
            buffers_iterable = (
                (f"{module_name}_buffer_{name}", buffer)
                for module_name, module in model.named_modules()
                for name, buffer in module.named_buffers()
                if isinstance(
                    module,
                    (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm),
                )
            )
            iterable = itertools.chain(iterable, buffers_iterable)
        return iterable

    def _save_current_model_state(self, model: nn.Module, model_state: Dict[str, Any]):
        """Copy the model's state to the provided dict."""
        for name, param in self.get_model_state_iterator(model):
            model_state[name] = param.detach().clone().to(device=self.device)

    def on_start(self, task: ClassyTask, local_variables: Dict[str, Any]) -> None:
        if self.state.model_state:
            # loaded state from checkpoint, do not re-initialize, only move the state
            # to the right device
            for name in self.state.model_state:
                self.state.model_state[name] = self.state.model_state[name].to(
                    device=self.device
                )
                self.state.ema_model_state[name] = self.state.ema_model_state[name].to(
                    device=self.device
                )
            return
        self._save_current_model_state(task.base_model, self.state.model_state)
        self._save_current_model_state(task.base_model, self.state.ema_model_state)

    def on_phase_start(self, task: ClassyTask, local_variables: Dict[str, Any]) -> None:
        # restore the right state depending on the phase type
        self.set_model_state(task, use_ema=not task.train)

    def on_phase_end(self, task: ClassyTask, local_variables: Dict[str, Any]) -> None:
        if task.train:
            # save the current model state since this will be overwritten by the ema
            # state in the test phase
            self._save_current_model_state(task.base_model, self.state.model_state)

    def on_update(self, task: ClassyTask, local_variables: Dict[str, Any]) -> None:
        with torch.no_grad():
            for name, param in self.get_model_state_iterator(task.base_model):
                self.state.ema_model_state[
                    name
                ] = self.decay * self.state.ema_model_state[name] + (
                    1 - self.decay
                ) * param.to(
                    device=self.device
                )

    def set_model_state(self, task: ClassyTask, use_ema: bool) -> None:
        """
        Depending on use_ema, set the appropriate state for the model.
        """
        model_state = self.state.ema_model_state if use_ema else self.state.model_state
        with torch.no_grad():
            for name, param in self.get_model_state_iterator(task.base_model):
                param.copy_(model_state[name])
